'''Date-related classes and functions'''

# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import sys
try:
    from DateTime import DateTime
except ImportError:
    pass # Zope is required

# ------------------------------------------------------------------------------
# Days of the week
weekDays = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
weekDays_ = weekDays + ('Off',)

# ------------------------------------------------------------------------------
def toUTC(d):
    '''When manipulating DateTime instances, like p_d, errors can raise when
       performing operations on dates that are not in Universal time, during
       months when changing from/to summer/winter hour. This function returns
       p_d set to UTC.'''
    return DateTime('%d/%d/%d UTC' % (d.year(), d.month(), d.day()))

# ------------------------------------------------------------------------------
class DayIterator:
    '''Class allowing to iterate over a range of days'''

    def __init__(self, startDay, endDay, back=False):
        self.start = toUTC(startDay)
        self.end = toUTC(endDay)
        # If p_back is True, the iterator will allow to browse days from end to
        # start.
        self.back = back
        self.finished = False
        # Store where we are within [start, end] (or [end, start] if back)
        if not back:
            self.current = self.start
        else:
            self.current = self.end

    def __iter__(self): return self
    def __next__(self):
        '''Returns the next day'''
        if self.finished:
            raise StopIteration
        res = self.current
        # Get the next day, forward
        if not self.back:
            if self.current >= self.end:
                self.finished = True
            else:
                self.current += 1
        # Get the next day, backward
        else:
            if self.current <= self.start:
                self.finished = True
            else:
                self.current -= 1
        return res
    next = __next__ # Python2-3 compliance

# ------------------------------------------------------------------------------
def getLastDayOfMonth(date, hour=None):
    '''Returns a DateTime object representing the last day of date.month()'''
    day = 31
    month = date.month()
    year = date.year()
    found = False
    while not found:
        try:
            res = DateTime('%d/%d/%d %s' % (year, month, day, hour or '12:00'))
            found = True
        except DateTime.DateError:
            day -= 1
    return res

def getDayInterval(date):
    '''Returns a tuple (startOfDay, endOfDay) representing the whole day into
       which p_date occurs.'''
    day = date.strftime('%Y/%m/%d')
    return DateTime('%s 00:00' % day), DateTime('%s 23:59' % day)

def getMonthInterval(date, hour=None):
    '''Returns a tuple (start, end) representing the start and end days for the
       month into which p_date is included.'''
    return DateTime(date.strftime('%Y/%m/01')), \
           getLastDayOfMonth(date, hour=hour)

def getSiblingMonth(date, next=True):
    '''Computes and returns a date corresponding to p_date but one month later
       (if p_next is True) or earlier (if p_next is False).'''
    if next:
        if date.month() == 12:
            year = date.year() + 1
            month = 1
        else:
            year = date.year()
            month = date.month() + 1
    else: # Get the previous month
        if date.month() == 1:
            year = date.year() - 1
            month = 12
        else:
            year = date.year()
            month = date.month() - 1
    month = str(month).zfill(2)
    fmt = '%d/%s/%%d %%H:%%M:%%S' % (year, month)
    dateStr = date.strftime(fmt)
    try:
        r = DateTime(dateStr)
    except Exception, e:
        # Start with the first day of the target month and get its last day
        fmt = '%d/%s/01' % (year, month)
        r = getLastDayOfMonth(DateTime(date.strftime(fmt)), hour='%H:%M:%S')
    return r

def periodsIntersect(start1, end1, start2, end2):
    '''Is there an intersection between intervals [start1, end1] and
       [start2, end2] ?'''
    # p_start1 and p_start2 must be DateTime instances.
    # p_end1 and p_end2 may be DateTime instances or None.
    # ~~~
    # Convert all parameters to seconds since the epoch
    if end1 is not None:
        end1 = end1._millis
    else:
        end1 = sys.maxint
    if end2 is not None:
        end2 = end2._millis
    else:
        end2 = sys.maxint
    start1 = start1._millis
    start2 = start2._millis
    # Intervals intersect if they are not disjoint
    if (start1 > end2) or (start2 > end1): return
    return True
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import copy
from persistent.list import PersistentList

from appy.px import Px
from appy.model.fields import Field
from appy.model.utils import Object as O
from appy.ui.layout import Layout, Layouts

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Total:
    '''Represents a computation that will be executed on a series of cells
       within a List field.'''
    def __init__(self, name, field, initValue):
        self.name = name # The sub-field name
        self.field = field # field.name is prefixed with the main field name
        self.value = initValue

    def __repr__(self):
        return '<Total %s=%s>' % (self.name, str(self.value))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Totals:
    '''If you want to add rows representing totals computed from regular List
       rows, specify it via Totals instances (see List attribute "totalRows"
       below).'''
    def __init__(self, id, label, subFields, onCell, initValue=0.0):
        # "id" must hold a short name or acronym and must be unique within all
        # Totals instances defined for a given List field.
        self.id = id
        # "label" is a i18n label that will be used to produce a longer name
        # that will be shown at the start of the total row.
        self.label = label
        # "subFields" is a list or tuple of sub-field names for which totals
        # must be computed.
        self.subFields = subFields
        # "onCell" stores a method that will be called every time a cell
        # corresponding to a field listed in self.subFields is walked in the
        # list. It will get these args:
        # * row      the current row as an Object instance;
        # * total    the Total instance (see above) corresponding to the current
        #            column;
        # * last     a boolean that is True if we are walking the last row.
        self.onCell = onCell
        # "initValue" is the initial value given to created Total instances
        self.initValue = initValue

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class List(Field):
    '''A list, stored as a list of Object instances ~[Object]~. Every object in
       the list has attributes named according to the sub-fields defined in this
       List.'''
    Totals = Totals

    # List is an "outer" field, made of inner fields
    outer = True

    class Layouts(Layouts):
        '''List-specific layouts'''
        g = Layouts(edit=Layout('f;rv=', width=None), view='fl')
        gd = Layouts(edit=Layout('d;v=f;', width=None), view='fl')
        # Default layouts for sub-fields
        sub = Layouts(Layout('frv', width=None))

        @classmethod
        def getDefault(class_, field):
            '''Default layouts for this List p_field'''
            return class_.g if field.inGrid() else super().getDefault(field)

    # A 1st invisible cell containing a virtual field allowing to detect this
    # row at server level.
    pxFirstCell = Px('''<td style="display: none">
     <table var="rid='%s*-row-*%s' % (field.name, rowId)"
            id=":'%d_%s' % (o.iid, rid)">
      <tr><td><input type="hidden" id=":rid" name=":rid"/></td></tr>
     </table></td>''')

    # PX for rendering a single row
    pxRow = Px('''
     <tr valign="top" style=":(rowId == -1) and 'display: none' or ''"
         class=":loop.row.odd and 'odd' or 'even'">
      <x if="not isCell">:field.pxFirstCell</x>
      <td if="showTotals"></td>
      <td for="info in subFields" if="info[1]"
          var2="x=field.updateTotals(totals, o, info, row, loop.row.last);
                field=info[1];
                minimal=isCell;
                fieldName='%s*%d' % (field.name, rowId)">:field.pxRender</td>
      <!-- Icons -->
      <td if="isEdit" align=":dright">
       <img class="clickable iconS" src=":url('deleteS.svg')"
            title=":_('object_delete')"
            onclick=":field.jsDeleteConfirm(q, tableId)"/>
       <img class="clickable iconS" src=":url('arrow.svg')"
            style="transform: rotate(180deg)" title=":_('move_up')"
            onclick=":'moveRow(%s,%s,this)' % (q('up'), q(tableId))"/>
       <img class="clickable iconS" src=":url('arrow.svg')"
            title=":_('move_down')"
            onclick=":'moveRow(%s,%s,this)' % (q('down'), q(tableId))"/>
       <input class="clickable" type="image" tabindex="0"
              src=":url('addBelow')" title=":_('object_add_below')"
              onclick=":'insertRow(%s,this); return false' % q(tableId)"/>
      </td></tr>''')

    # Display totals (on view only) when defined
    pxTotals = Px('''
     <tr for="totalRow in field.totalRows" var2="total=totals[totalRow.id]">
      <th>:_(totalRow.label)</th>
      <th for="info in subFields">:field.getTotal(total, info)</th>
     </tr>''')

    # PX for rendering the list (shared between pxView, pxEdit and pxCell)
    pxTable = Px('''
     <table var="isEdit=layout == 'edit';
                 isCell=layout == 'cell';
                 tableId='list_%s' % name" if="isEdit or value"
            id=":tableId" width=":field.width" class="grid"
            var2="subFields=field.getSubFields(o, layout);
                  totals=field.createTotals(isEdit);
                  showTotals=not isEdit and totals">
      <!-- Header -->
      <thead>
       <tr valign="middle">
        <th if="showTotals"></th>
        <th for="name, sub in subFields" if="sub"
            width=":field.widths[loop.name.nb]">
         <x var="field=sub">::field.getListHeader(_ctx_)</x></th>
        <!-- Icon for adding a new row -->
        <th if="isEdit">
         <img class="clickable" src=":url('plus')" title=":_('object_add')"
              onclick=":'insertRow(%s)' % q(tableId)"/>
        </th>
       </tr>
      </thead>

      <!-- Template row (edit only) -->
      <x var="rowId=-1" if="isEdit">:field.pxRow</x>

      <!-- Rows of data -->
      <x var="rows=field.getInputValue(inRequest, requestValue, value)"
         for="row in rows" var2="rowId=loop.row.nb">:field.pxRow</x>

      <!-- Totals -->
      <x if="showTotals">:field.pxTotals</x>
     </table>''')

    view = cell = Px('''<x>:field.pxTable</x>''')
    edit = Px('''<x>
     <!-- This input makes Appy aware that this field is in the request -->
     <input type="hidden" name=":name" value=""/><x>:field.pxTable</x>
    </x>''',

     js='''
       updateRowNumber = function(row, rowIndex, action) {
         /* Within p_row, we must find tables representing fields. Every such
            table has an id of the form [objectId]_[field]*[subField]*[i].
            Within this table, for every occurrence of this string, the "[i]"
            must be replaced with an updated index. If p_action is 'set',
            p_rowIndex is this index. If p_action is 'add', the index must
            become [i] + p_rowIndex. */

         // Browse tables representing fields
         var fields = row.getElementsByTagName('table'),
             tagTypes = ['input', 'select', 'img', 'textarea', 'a', 'script'],
             newIndex = -1, id, old, elems, neww, val, w, oldIndex;

         // Patch fields
         for (var i=0; i<fields.length; i++) {
           // Extract, from the table ID, the field identifier
           id = fields[i].id;
           if ((!id) || (id.indexOf('_') == -1)) continue;
           // Extract info from the field identifier
           old = id.split('_')[1];
           elems = old.split('*');
           /* Get "old" as a regular expression: we may need multiple
              replacements. */
           old = new RegExp(old.replace(/\\*/g, '\\\\*'), 'g');
           // Compute the new index (if not already done) and new field ID
           if (newIndex == -1) {
             oldIndex = parseInt(elems[2]);
             newIndex = (action == 'set')? rowIndex: oldIndex + rowIndex;
           }
           neww = elems[0] + '*' + elems[1] + '*' + newIndex;
           // Replace the table ID with its new ID
           fields[i].id = fields[i].id.replace(old, neww);
           // Find sub-elements mentioning "old" and replace it with "neww"
           val = w = null;
           for (var j=0; j<tagTypes.length; j++) {
             var widgets = fields[i].getElementsByTagName(tagTypes[j]);
             for (var k=0; k<widgets.length; k++) {
               w = widgets[k];
               // Patch id
               val = w.id;
               if (val) w.id = val.replace(old, neww);
               // Patch name
               val = w.name;
               if (val) w.name = val.replace(old, neww);
               // Patch href
               if ((w.nodeName == 'A') && w.href)
                 { w.href = w.href.replace(old, neww); }
               // Patch (and reeval) script
               if (w.nodeName == 'SCRIPT') {
                 w.text = w.text.replace(old, neww);
                 eval(w.text);
               }
             }
           }
         }
       }

       insertRow = function(tableId, previous) {
         /* Add a new row in table with ID p_tableId, after p_previous row or at
            the end if p_previous is null. */
         var table = document.getElementById(tableId),
             body = table.tBodies[0],
             rows = table.rows,
             newRow = rows[1].cloneNode(true),
             next = (previous)? \
                    previous.parentNode.parentNode.nextElementSibling : null;
         newRow.style.display = 'table-row';
         if (next) {
           var newIndex = next.rowIndex;
           // We must insert the new row before it
           body.insertBefore(newRow, next);
           // The new row takes the index of "next" (- the 2 unsignificant rows)
           updateRowNumber(newRow, newIndex-2, 'set');
           // Row numbers for "next" and the following rows must be incremented
           for (var i=newIndex+1; i < rows.length; i++) {
             updateRowNumber(rows[i], 1, 'add');
           }
         }
         else {
           // We must insert the new row at the end
           body.appendChild(newRow);
           // Within newRow, incorporate the row nb within field names and ids
           updateRowNumber(newRow, rows.length-3, 'set');
         }
       }

       deleteRow = function(tableId, deleteImg, ask, rowIndex) {
         var table = document.getElementById(tableId),
             rows = table.rows,
             row = (deleteImg)? deleteImg.parentNode.parentNode: rows[rowIndex];
             rowIndex = row.rowIndex;
         // Must we ask the user to confirm this action ?
         if (ask) {
             askConfirm('script',
                        'deleteRow("'+tableId+'",null,false,'+rowIndex+')');
             return;
         }
         // Decrement higher row numbers by 1 because of the deletion
         for (var i=rowIndex+1; i < rows.length; i++) {
           updateRowNumber(rows[i], -1, 'add');
         }
         table.deleteRow(rowIndex);
       }

       moveRow = function(direction, tableId, moveImg) {
         var row = moveImg.parentNode.parentNode,
             body = document.getElementById(tableId).tBodies[0], sibling;
         // Move up or down
         if (direction == 'up') {
           sibling = row.previousElementSibling;
           if (sibling && sibling.style.display != 'none') {
             updateRowNumber(row, -1, 'add');
             updateRowNumber(sibling, 1, 'add');
             body.insertBefore(row, sibling);
           }
         }
         else if (direction == 'down') {
           sibling = row.nextElementSibling;
           if (sibling) {
             updateRowNumber(row, 1, 'add');
             updateRowNumber(sibling, -1, 'add');
             // If sibling is null, row will be added to the end
             sibling = sibling.nextElementSibling;
             body.insertBefore(row, sibling);
           }
         }
       }''')

    search = ''

    def __init__(self, fields, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, readPermission='read', writePermission='write', width='',
      height=None, maxChars=None, colspan=1, master=None, masterValue=None,
      focus=False, historized=False, mapping=None, generateLabel=None,
      label=None, subLayouts=Layouts.sub, widths=None, view=None, cell=None,
      edit=None, xml=None, translations=None, deleteConfirm=False,
      totalRows=None):
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
         show, page, group, layouts, move, False, True, None, None, False, None,
         readPermission, writePermission, width, height, None, colspan, master,
         masterValue, focus, historized, mapping, generateLabel, label, None,
         None, None, None, True, False, view, cell, edit, xml, translations)
        self.validable = True
        # Tuple of elements of the form (name, Field instance) determining the
        # format of every element in the list.
        self.fields = fields
        # Force some layouting for sub-fields, if subLayouts are given. So the
        # one who wants freedom on tuning layouts at the field level must
        # specify subLayouts=None.
        if subLayouts:
            for name, field in self.fields:
                field.layouts = subLayouts
        # One may specify the width of every column in the list. Indeed, using
        # widths and layouts of sub-fields may not be sufficient.
        self.computeWidths(widths)
        # When deleting a row, must we display a popup for confirming it ?
        self.deleteConfirm = deleteConfirm
        # If you want to specify additional rows representing totals, give in
        # "totalRows" a list of Totals instances (see above).
        self.totalRows = totalRows or []

    def init(self, class_, name):
        '''List-specific lazy initialisation'''
        Field.init(self, class_, name)
        for sub, field in self.fields:
            fullName = '%s_%s' % (name, sub)
            field.init(class_, fullName)
            field.name = '%s*%s' % (name, sub)

    def computeWidths(self, widths):
        '''Set given p_widths or compute default ones if not given'''
        self.widths = widths or ([''] * len(self.fields))

    def getField(self, name):
        '''Gets the field definition whose name is p_name'''
        for n, field in self.fields:
            if n == name: return field

    def getSubFields(self, o, layout):
        '''Returns the sub-fields (name, Field) that are showable among
           field.fields on the given p_layout. Fields that cannot appear in
           the result are nevertheless present as a tuple (name, None). This
           way, it keeps a nice layouting of the table.'''
        r = []
        for n, field in self.fields:
            elem = (n, None)
            if field.isShowable(o, layout):
                elem = (n, field)
            r.append(elem)
        return r

    def getRequestValue(self, o, requestName=None):
        '''Concatenates the list from distinct form elements in the request'''
        req = o.req
        name = requestName or self.name # A List may be into another List (?)
        prefix = name + '*-row-*' # Allows to detect a row of data for this List
        r = {}
        isDict = True # We manage both List and Dict
        for key in req.keys():
            if not key.startswith(prefix): continue
            # I have found a row: get its index
            row = O()
            rowId = key.split('*')[-1]
            if rowId == '-1': continue # Ignore the template row
            for subName, subField in self.fields:
                keyName = '%s*%s*%s' % (name, subName, rowId)
                if keyName + subField.getRequestSuffix() in req:
                    v = subField.getRequestValue(o, requestName=keyName)
                    setattr(row, subName, v)
            if rowId.isdigit():
                rowId = int(rowId)
                isDict = False
            r[rowId] = row
        # Produce a sorted list (List only)
        if not isDict:
            keys = list(r.keys())
            keys.sort()
            r = [r[key] for key in keys]
        # I store in the request this computed value. This way, when individual
        # subFields will need to get their value, they will take it from here,
        # instead of taking it from the specific request key. Indeed, specific
        # request keys contain row indexes that may be wrong after row deletions
        # by the user.
        if r: req[name] = r
        return r

    def setRequestValue(self, o):
        '''Sets in the request, the field value on p_o in its "request-
           carriable" form.'''
        value = self.getValue(o)
        if value is not None:
            req = o.req
            name = self.name
            i = 0
            for row in value:
                req['%s*-row-*%d' % (name, i)] = ''
                for n, v in row.d().items():
                    key = '%s*%s*%d' % (name, n, i)
                    req[key] = v
                i += 1

    def getStorableRowValue(self, o, requestValue):
        '''Gets a ready-to-store Object instance representing a single row,
           from p_requestValue.'''
        r = O()
        for name, field in self.fields:
            if not hasattr(requestValue, name) and \
               not field.isShowable(o, 'edit'):
                # Some fields, although present on "edit", may not be part of
                # the p_requestValue (ie, a "select multiple" with no value at
                # all will not be part of the POST HTTP request). If we want to
                # know if the field was in the request or not, we must then
                # check if it is showable on layout "edit".
                continue
            subValue = getattr(requestValue, name, None)
            try:
                setattr(r, name, field.getStorableValue(o, subValue))
            except ValueError:
                # The value for this field for this specific row is incorrect.
                # It can happen in the process of validating the whole List
                # field (a call to m_getStorableValue occurs at this time). We
                # don't care about it, because later on we will have sub-field
                # specific validation that will also detect the error and will
                # prevent storing the wrong value in the database.
                setattr(r, name, subValue)
        return r

    def getStorableValue(self, o, value, complete=False):
        '''Gets p_value in a form that can be stored in the database'''
        return [self.getStorableRowValue(o, v) for v in value]

    def getCopyValue(self, o):
        '''Return a (deep) copy of field value on p_obj''' 
        r = getattr(o, self.name, None)
        if r: return copy.deepcopy(r)

    def getCss(self, layout, r):
        '''Gets the CSS required by sub-fields if any'''
        for name, field in self.fields:
            field.getCss(layout, r)

    def getJs(self, layout, r, config):
        '''Gets the JS required by sub-fields if any'''
        for name, field in self.fields:
            field.getJs(layout, r, config)

    def jsDeleteConfirm(self, q, tableId):
        '''Gets the JS code to call when the user wants to delete a row'''
        confirm = 'true' if self.deleteConfirm else 'false'
        return 'deleteRow(%s,this,%s)' % (q(tableId), confirm)

    def subValidate(self, o, value, errors):
        '''Validates inner fields'''
        i = -1
        for row in value:
            i += 1
            for name, subField in self.fields:
                message = subField.validate(o, getattr(row, name, None))
                if message:
                    setattr(errors, '%s*%d' % (subField.name, i), message)

    def createTotals(self, isEdit):
        '''When rendering the List field, if total rows are defined, create a
           Total instance for every sub-field for which a total must be
           computed.'''
        if isEdit or not self.totalRows: return
        r = {} # Keyed by Totals.name
        for totals in self.totalRows:
            subTotals = O()
            for name in totals.subFields:
                totalObj = Total(name, self.getField(name), totals.initValue)
                setattr(subTotals, name, totalObj)
            r[totals.id] = subTotals
        return r

    def updateTotals(self, totals, o, info, row, last):
        '''Every time a cell is encountered while rendering the List, this
           method is called to update totals when needed'''
        if not totals: return
        # Browse Totals instances
        for totalRow in self.totalRows:
            # Are there totals to update ?
            total = getattr(totals[totalRow.id], info[0], None)
            if total:
                totalRow.onCell(o, row, total, last)

    def getTotal(self, totals, info):
        '''Get the total for the field p_info if available'''
        total = getattr(totals, info[0], None)
        if total: return total.value
        return ''
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import xml.sax, io
from xml.sax.xmlreader import InputSource
from html.entities import entitydefs as htmlEntities
from xml.parsers.expat import XML_PARAM_ENTITY_PARSING_NEVER
from xml.sax.handler import ContentHandler, ErrorHandler, feature_external_ges

# Constants  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
xmlPrologue = '<?xml version="1.0" encoding="utf-8" ?>\n'
xhtmlPrologue = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" '\
                '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n'

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Tag:
    '''Represents an XML tag'''

    def __init__(self, name, attrs=None, nsUri=None, parent=None):
        '''A Tag instance may represent:
           - an already parsed tag (in this case, p_name may be prefixed with a
             namespace);
           - the definition of an XML tag (in this case, no namespace can be
             found in p_name; but a namespace URI may be defined in p_nsUri).'''
        self.name = name
        self.attrs = attrs
        if name.find(':') != -1:
            self.nsName = name
            self.ns, self.name = name.split(':', 1)
        else:
            self.nsName = self.name = name
            self.ns = ''
            self.nsUri = nsUri
        self.parent = parent # The parent element

    def __repr__(self):
        r = '<tag %s' % self.name
        if self.attrs:
            attrs = ['%s="%s"' % (k, v) for k, v in self.attrs.items()]
            r += ' [%s]' % ', '.join(attrs)
        return '%s>' % r

    def getFullName(self, namespaces=None):
        '''Gets the name of the element including the namespace prefix'''
        if not namespaces:
            r = self.name
        else:
            r = '%s:%s' % (namespaces[self.nsUri], self.name)
        return r

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Environment:
    '''An XML environment remembers a series of elements during a SAX parsing.
       This class is an abstract class that gathers basic things like
       namespaces.'''

    def __init__(self):
        '''Initialises an XML environment'''
        # Dict "namespaces" contains the XML namespace declarations encountered
        # so far.
        self.namespaces = {} # ~{s_namespaceUri: s_namespaceName}~
        self.currentTag = None # The currently parsed tag
        self.parser = None

    def manageNamespaces(self, attrs):
        '''Manages namespaces definitions encountered in p_attrs'''
        for name, value in attrs.items():
            if name.startswith('xmlns:'):
                self.namespaces[value] = name[6:]

    def ns(self, nsUri):
        '''Returns the namespace corresponding to o_nsUri'''
        return self.namespaces[nsUri]

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Parser(ContentHandler, ErrorHandler):
    '''Basic expat-based XML parser'''

    # This parser manages:
    # - the stack of currently parsed tags;
    # - namespace declarations.
    # It knows about HTML entities.

    def __init__(self, env=None, caller=None, raiseOnError=True):
        '''p_env should be an instance of a class that inherits from
           appy.xml.Environment: it specifies the environment to use for this
           SAX parser.'''
        ContentHandler.__init__(self)
        self.env = env or Environment()
        self.env.parser = self
        # The class calling this parser
        self.caller = caller
        # Fast, standard expat parser
        self.parser = xml.sax.make_parser()
        # The result of parsing
        self.r = None
        # Raise or not an error when a parsing error is encountered
        self.raiseOnError = raiseOnError

    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # ContentHandler methods
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def startDocument(self):
        parser = self.parser._parser
        parser.UseForeignDTD(True)
        parser.SetParamEntityParsing(XML_PARAM_ENTITY_PARSING_NEVER)

    def setDocumentLocator(self, locator):
        self.locator = locator
        return self.env

    def endDocument(self): return self.env

    def startElement(self, name, attrs):
        env = self.env
        env.manageNamespaces(attrs)
        # Replace the current elem (or not if this is the root tag) by a fresh
        # Tag instance.
        env.currentTag = Tag(name, attrs=attrs, parent=env.currentTag)
        return env

    def endElement(self, name):
        '''Pops the currently walked elem'''
        env = self.env
        env.currentTag = env.currentTag.parent
        return env

    def characters(self, content): return self.env

    def skippedEntity(self, entity):
        '''Called every time expat does not recognize an entity. We provide here
           support for HTML entities.'''
        # When an unknown entity is encountered, put a question mark in the
        # result instead of raising an exception.
        char = htmlEntities.get(entity) or '?'
        self.characters(char)

    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # ErrorHandler methods
    #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def error(self, error):
        if self.raiseOnError: raise error
        else: print('SAX error %s' % str(error))

    def fatalError(self, error):
        if self.raiseOnError: raise error
        else:
            print('SAX fatal error %s on XML:' % str(error))

    def warning(self, error): pass

    def parse(self, xml, source='string'):
        '''Parses a XML stream'''
        # ----------------------------------------------------------------------
        # If p_source | p_xml...
        # is...       |
        # ----------------------------------------------------------------------
        #  "string"   | must be a string or bytes object containing valid XML
        #             | content;
        # ----------------------------------------------------------------------
        #  "file"     | can be:
        #             |  - a string containing the path to the XML file on disk;
        #             |  - a file handler opened for reading. Note that in this
        #             |    case, this method will close it.
        # ----------------------------------------------------------------------
        self._xml = xml
        self.parser.setContentHandler(self)
        self.parser.setErrorHandler(self)
        self.parser.setFeature(feature_external_ges, False)
        inputSource = InputSource()
        if source == 'string':
            xml = xml.decode() if isinstance(xml, bytes) else xml
            inputSource.setByteStream(io.StringIO(xml))
        else:
            xml = xml if isinstance(xml, io.IOBase) else open(xml)
            inputSource.setByteStream(xml)
        self.parser.parse(inputSource)
        if isinstance(xml, io.IOBase): xml.close()
        return self.r
#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This script allows to generate a Zope product from a Appy application.'''

# ------------------------------------------------------------------------------
import sys, os.path
from optparse import OptionParser
from appy.gen.generator import GeneratorError, ZopeGenerator
from appy.shared.utils import LinesCounter
from appy.shared.packaging import Debianizer
import appy.version

# ------------------------------------------------------------------------------
ERROR_CODE = 1
APP_NOT_FOUND = 'Application not found at %s.'
WRONG_NG_OF_ARGS = 'Wrong number of arguments.'
C_OPTION = 'Removes from i18n files all labels that are not automatically ' \
           'generated from your gen-application. It can be useful during ' \
           'development, when you do lots of name changes (classes, ' \
           'attributes, states, transitions, etc): in this case, the Appy ' \
           'i18n label generation machinery produces lots of labels that ' \
           'then become obsolete.'
D_OPTION = 'Generates a Debian package for this app. The Debian package will ' \
           'be generated at the same level as the root application folder.'

class GeneratorScript:
    '''usage: %prog [options] app

       "app"          is the path to your Appy application, which must be a
                      Python package (= a folder containing a file named
                      __init__.py). Your app may reside anywhere, but needs to
                      be accessible by Zope. Typically, it may be or symlinked
                      in <yourZopeInstance>/lib/python.

       This command generates a Zope product in <app>/zope, which must be
       or symlinked in <yourZopeInstance>/Products.
    '''
    def manageArgs(self, parser, options, args):
        # Check number of args
        if len(args) != 1:
            print(WRONG_NG_OF_ARGS)
            parser.print_help()
            sys.exit(ERROR_CODE)
        # Check existence of application
        if not os.path.exists(args[0]):
            print(APP_NOT_FOUND % args[0])
            sys.exit(ERROR_CODE)
        # Convert app path to an absolute path
        args[0] = os.path.abspath(args[0])

    def run(self):
        optParser = OptionParser(usage=GeneratorScript.__doc__)
        optParser.add_option("-c", "--i18n-clean", action='store_true',
            dest='i18nClean', default=False, help=C_OPTION)
        optParser.add_option("-d", "--debian", action='store_true',
            dest='debian', default=False, help=D_OPTION)
        (options, args) = optParser.parse_args()
        try:
            self.manageArgs(optParser, options, args)
            print('Appy version: %s' % appy.version.verbose)
            print('Generating Zope product in %s/zope...' % args[0])
            ZopeGenerator(args[0], options).run()
            # Give the user some statistics about its code
            LinesCounter(args[0], excludes=['%szope' % os.sep]).run()
            # Generates a Debian package for this app if required
            if options.debian:
                app = args[0]
                appDir = os.path.dirname(app)
                appName = os.path.basename(app)
                # Get the app version from zope/version.txt
                f = file(os.path.join(app, 'zope', appName, 'version.txt'))
                version = f.read()
                f.close()
                version = version[:version.find('build')-1]
                Debianizer(app, appDir, appVersion=version).run()
        except GeneratorError, ge:
            sys.stderr.write(str(ge))
            sys.stderr.write('\n')
            optParser.print_help()
            sys.exit(ERROR_CODE)

# ------------------------------------------------------------------------------
if __name__ == '__main__':
    GeneratorScript().run()
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>Champ FCK</p>
<ol>
    <li>aaaa
    <ol>
        <li>Azerty</li>
        <li>aaaa</li>
    </ol>
    </li>
    <li>edzfrgh</li>
    <li>Kupu</li>
</ol>
<table cellspacing="1" cellpadding="1" border="1" style="width: 210px; height: 66px;">
    <tbody>
        <tr>
            <td>a</td>
            <td>b</td>
        </tr>
        <tr>
            <td>x</td>
            <td>vvv</td>
        </tr>
        <tr>
            <td>bbb</td>
            <td>vvvvv</td>
        </tr>
    </tbody>
</table>
<p>&nbsp;</p>
<p style="margin-left: 40px;">hghdghghgh</p>
<p>With non-breaking dashes: Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, 
Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, 
Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, 
Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, Jean‑Michel, 
</p>
<p>With normal dashes: Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel,
Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, 
Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel, Jean-Michel
</p>
<ul>
    <li>aaa</li>
    <li>&nbsp;</li>
    <li>bvbb</li>
</ul>
<ol>
    <li>regrg</li>
    <li>&nbsp;</li>
</ol>
<p>vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù&nbsp; vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù</p>
<p>&nbsp;</p>
<p>vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù&nbsp; vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@èù vghriqghrghgfd&nbsp; hgkll hgjkf lghjfkd slhgjfd klhgjfds klghjfds s&amp;é@</p>
'''



'''The scheduler manages cron-like actions being programmed and automatically
   executed by the Appy server, as "virtual" requests.'''

# Important notes  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#
# I. The scheduler runs in "best effort" mode. Everytime a series of jobs must
#    be run according to the current time, they are ran, sequentially, in the
#    order of their definition. If their execution takes several minutes, during
#    this period, the scheduler will not trigger any other job.

#    Example
#    The scheduler is configured with 3 jobs, in that order:
#    - job A must be run at midnight     (cron syntax: 0   0 * * *) ;
#    - job B must be run every 3 minutes (cron syntax: */3 * * * *) ;
#    - job C must be run at 00:15        (cron syntax: 15  0 * * *).

#    At midnight, the scheduler determines that both jobs A and B must be run.
#    - job A is ran at 00:00 and takes 34 minutes to execute ;
#    - job B is ran at 00:34 and takes 30 seconds ;
#    - job B's next execution is at 00:36, then at 00:39, etc.

#    Analysis
#    a) The first execution of job B didn't occur at 00:00, neither at 00:03,
#       but after job A has completely terminated.
#    b) From 00:00 to 00:34, job B was not executed, although, according to its
#       timedef, it should have run every 3 minutes during this period.
#    c) Job C was not executed at all. Indeed, at 00:15, the scheduler was busy
#       with the execution of job A and was blocked, preventing him to detect
#       any other job execution. Job C's next execution will be the day after,
#       at 00:15, excepted if, again, execution of jobs A + B takes more than 15
#       minutes.
#    d) When the scheduler "wakes up" at 00:34, after having executed job A, he
#       does not "look backwards" to check whether other jobs should have been
#       run in the meanwhile, executing job C once, and job B as many times as
#       required between 00:00 and 00:34. No. Instead, he executes the other job
#       (job B) tha was also planned at midnight, he ignores everything
#       scheduled between 00:00 and 00:34 and continues to check for jobs,
#       starting at 00:35.

# II. The scheduler delays any incoming HTTP request. While jobs are executed by
#     the scheduler, the infinite server loop, used by the server to check for
#     incoming HTTP requests, is paused. The next HTTP request will be handled
#     after all currently running jobs have ended.

#     So, it is preferable to "size" your jobs into a series of "short-term"
#     jobs, triggered at regular intervals, rather than defining "long-lasting"
#     jobs, excepted if you are sure these latter are run when no web user is
#     there, ie, during the night.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import time

from appy.server.handler import VirtualHandler

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
MIN_KO   = 'Config attribute "jobs.minutes" must be an integer being higher ' \
           'or equal to 1.'
MISSING  = 'Missing %s for a job.'
WRONG_TD = 'Wrong timedef "%s".'
TDEF_KO  = '%s. Must be of the form "m h dom mon dow".' % WRONG_TD

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Job:
    '''Represents a job that must be executed in the context of the scheduler'''

    def __init__(self, method):
        # Running a job consists in executing a method as defined on the tool.
        # m_method is the name of this method.
        method = method.strip() if method else method
        if not method: raise Exception(MISSING % 'method')
        self.method = method
        # This instance will also hold more information, stored by the Appy
        # server itself. For example, if this job is scheduled at regular
        # intervals, the date of the last execution will be stored on this Job
        # instance. Consequently, do not reuse the same job instance for several
        # config entries. In order to avoid name clashes, any attribute added by
        # Appy will start with an underscore.
        # ~~~
        # By default, the job will lead to a database commit. As for any other
        # method executed in the context of a UI request, if m_method must not
        # lead to a commit (either because the job, intrinsically, dos not
        # update any data in the database, or because an error occurred), the
        # method can access the currently running handler and set its "commit"
        # attribute, ie:
        # 
        #                      tool.H().commit = False

    def run(self, scheduler):
        '''Runs this job'''
        server = scheduler.server
        # Create a Virtual handler for that
        handler = VirtualHandler(server)
        # Run the job
        server.database.run(self.method, handler, scheduler.logger)
        # Unregister the virtual handler
        VirtualHandler.remove()

    def __repr__(self):
        '''p_self's string representation'''
        return '<Job %s>' % self.method

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TimePart:
    '''Appy representation of a "time part", being any of the 5 parts of a time
       def (see class TimeDef).'''

    # Note that time part "*" will not be represented by a TimePart instance,
    # but as an entry whose value is None in dict TimeDef.parts.

    def __init__(self, part):
        # The "base" of a p_part can be:
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # # |   Base      | Meaning: the part matches ...
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # 1 |     *       | any value for the corresponding measured element
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # 2 |   <int>     | if the value for the corresponding measured element
        #   |             | equals this number.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # 3 | <min>-<max> | if the value for the corresponding measured element
        #   |             | is within the [min, max] range of values.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        # The "recurrence" of a p_part restricts the values matched by the base,
        # if this latter represents a set of values (cases #1 and #3 hereabove).
        # Recurrence is represented by suffix "/<int>" and means
        # "every >int> units". Here are some examples.
        #
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #    Example    | Meaning
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #      */3      | Every 3 units (0, 3, 6, 9 ...)
        #    10-20/3    | Every 3 units, between 10 and 20: 10, 13, 16, 19.
        #      1/2      | Illegal: recurrence can only restrict a base
        #               | representing a set of values, not a single value.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        # You may say: what about notation 5,8 for example ? Such situations are
        # managed by a list of TimePart instances. In the example, 2 TimePart
        # instances would have be created, with bases 5 and 8, respectively.
        self.part = part

        # p_part will be broken down into these elements
        self.numA = None
        self.numB = None
        self.rec = None

        # Here is the correspondence between a p_part and its components
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #    Base     | Attribute values
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #      *      | numA is None,  numB is None
        #    <int>    | numA is <int>, numB is None
        # <min>-<max> | numA is <min>, numB is <max>
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Attribute "rec" will be None if no recurrency is defined, or will
        # store the specified recurrence, as an integer number.
        self.parsePart()

    def parsePart(self):
        '''Parses p_part and get its components'''
        part = self.part
        if part == '*': return
        splitted = part.split('/', 1)
        # Manage the recurrence
        if len(splitted) == 2:
            # A "rec" part is present
            self.rec = int(splitted[1])
        # Manage the base
        base = splitted[0]
        if base == '*':
            pass # Nothing more to do
        elif base.isdigit():
            self.numA = int(base)
        else:
            A, B = base.split('-')
            self.numA = int(A)
            self.numB = int(B)

    def matches(self, value):
        '''Does this time part (p_self) match this p_value ?'''
        noneA = self.numA is None
        noneB = self.numB is None
        noneRec = self.rec is None
        # A single value
        if not noneA and noneB:
            return value == self.numA
        # Case: *
        elif noneA and noneB:
            if noneRec: return True
            # Take recurrence into account
            return value % self.rec == 0
        # An interval of values
        elif not noneA and not noneB:
            # The p_value must be included in it
            if value < self.numA or value > self.numB: return
            if noneRec: return True
            # Take the recurrence into account
            return (value - self.numA) % self.rec == 0

    def __repr__(self):
        '''p_self's string representation'''
        rec = self.rec and ('/%d' % self.rec) or ''
        numA = self.numA
        numB = self.numB
        noneA = numA is None
        noneB = numB is None
        if noneA and noneB:
            r = '*'
        elif not noneA and not noneB:
            r = '%d-%d' % (numA, numB)
        else:
            r = str(numA)
        return '<TimePart %s%s>' % (r, rec)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TimeDef:
    '''Internal Appy representation of a "time definition" as specified by the
       cron syntax.'''

    # A TimeDef instance represents a cron entry, whose "task" part has been
    # removed. Indeed, this "task" part is represented by class Job hereabove.

    # For example, cron entry
    #
    #                         30 0 * * * /my/script.sh
    #
    # is represented in the Appy scheduler as a tuple
    #
    #               (TimeDef('30 0 * * *'), Job('someToolMethod'))

    # Constants representing the 5 time parts in a timedef
    MM  = 0 # Minutes      (0-59)
    HH  = 1 # Hours        (0-23)
    dd  = 2 # Day of month (1-31)
    mm  = 3 # Month        (1-12)
    dow = 4 # Day of week  (0-6 = Sunday to Saturday; 7=also Sunday)

    # Shortcuts: predefined values, starting with @
    shortcuts = {
     'yearly'  : '0 0 1 1 *',
     'annually': '0 0 1 1 *',
     'monthly' : '0 0 1 * *',
     'weekly'  : '0 0 * * 0',
     'daily'   : '0 0 * * *',
     'midnight': '0 0 * * *',
     'hourly'  : '0 * * * *'
    }

    def __init__(self, timeDef):
        self.timeDef = timeDef.strip() if timeDef else None
        # The timedef will be stored as a dict ~{i_partID: None|[TimePart]}~
        # * Every key is one constant among MM, HH, dd, mm and dow as defined on
        #   this class.
        # * Every value can be one of the following.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #   Value    | Meaning
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #    None    | It corresponds to a time part being the star, "*"
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # [TimePart] | For any other value, a list of TimePart instances will be
        #            | created (see hereabove). If the time part includes
        #            | several comma-separated values, there will be as many
        #            | TimePart instances. Else, the list will contain a single
        #            | TimePart instance.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        self.parts = {}
        self.parse()

    def parse(self):
        '''Parses p_self.timeDef'''
        # Ensure the timedef is not empty
        tdef = self.timeDef
        if not tdef: raise Exception(MISSING % 'timeDef')
        # Convert shortcuts
        if tdef.startswith('@'):
            tdef = TimeDef.shortcuts[tdef[1:]]
        # Split the timedef into parts
        sparts = tdef.split()
        # Ensure it includes 5 time parts
        if len(sparts) != 5:
            raise Exception(TDEF_KO % timeDef)
        # Store the parts as TimePart instances in dict p_self.parts
        i = 4
        parts = self.parts
        while i >= 0:
            # Get the "string" part
            spart = sparts[i]
            # Convert it to its appropriate internal structure
            if spart == '*':
                value = None
            else:
                value = []
                for sval in spart.split(','):
                    value.append(TimePart(sval))
            parts[i] = value
            i -= 1

    # Mapping between the parts of a timedef and the names of the corresponding
    # parts in the struct returned by time.localtime():
    #
    # - each key is the index of the value in the timedef ;
    # - each value is a tuple (j, delta, ifNegative):
    #
    #   * "j" is the index of the corresponding part in the data structure
    #     returned by time.localtime() ;
    #
    #   * "delta" is an optional value to add to the value as returned by
    #     time.localtime(). Indeed, Python values may differ from those defined
    #     by the crontab syntax. For example, according to Python, 0 is the
    #     number for Monday; according to cron, it is the number for Sunday
    #     (together with 7). By applying a delta of -1 to the Python value, a
    #     correct comparison may de done ;
    #
    #   * "ifNegative" is the number to add to the Python value, if negative,
    #     after the "delta" has been applied to it. In the previous example, if
    #     0 (Sunday) is specified, applying -1 produces -1, being negative.
    #     Adding 7 produces the correct corresponding Python value.
    codeMap = {
      dow: (6, -1, 7), # Day of the week - Monday is 0 (Python) or 1 (cron)
      mm:  (1,  0, 0), # Month number    - from 1 to 12
      dd:  (2,  0, 0), # Day number      - from 1 to 31
      HH:  (3,  0, 0), # Hour            - from 0 to 23
      MM:  (4,  0, 0)  # Minutes         - from 0 to 59
    }

    def mustRun(self, now):
        '''Does this timedef (p_self) match p_now ?'''
        # Walk parts from the last (dow) to the first part (minutes)
        parts = self.parts
        for i, info in TimeDef.codeMap.items():
            timeParts = parts[i]
            # A part being None (= *) always matches
            if timeParts is None: continue
            # Get the current time value, calibrated
            j, delta, ndelta = info
            fig = now[j] + delta
            if fig < 0:
                fig += ndelta
            # Does this time value match the time part(s) ?
            matches = False
            for timePart in timeParts:
                if timePart.matches(fig):
                    matches = True
                    break
            if not matches:
                return
        # If we are here, this timedef matches p_now
        return True

    def __repr__(self):
        '''p_self's string representation'''
        return '<TimeDef %s>' % self.parts

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Scheduler configuration, defining the list of planned jobs.'''

    def __init__(self):
        # The scheduler will be run every "minutes" minutes. It is not advised
        # to change this value, at the risk of being unable to conform to the
        # crontab semantics.
        # ~~~
        # For example, if you decide to set the value to 5, every timeDef entry
        # where the "minutes" part is not a multiple of 5 will never match.
        self.minutes = 1
        # Attribute "all" lists all the planned jobs. Every entry in this list
        # must be a tuple (TimeDef, Job).
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # timeDef | Refers to an instance of class TimeDef hereabove.
        #         |
        #         | TimeDef conforms to crontab syntax. In the future, it will
        #         | be extended to incorporate load-related definitions. The
        #         | goal is to implement behaviours like this one: "execute this
        #         | job if the system load was low during the last minutes".
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Job     | A Job instance (see class Job hereabove), defining the
        #         | method to execute.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Use method m_add below to add a job to this list.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        self.all = []

    def check(self):
        '''Ensures the Scheduler config is correct'''
        # Check attribute "minutes"
        minutes = self.minutes
        if not isinstance(minutes, int) or (minutes < 1):
            raise Exception(MIN_KO)

    def add(self, timeDef, method):
        '''Adds a job to p_self.all, for running this m_method according to this
           p_timeDef.'''
        self.all.append( (TimeDef(timeDef), Job(method)) )

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Scheduler:
    '''The Appy scheduler'''

    def __init__(self, server):
        '''A unique Scheduler instance is created per Appy server'''
        # The Server instance
        self.server = server
        self.logger = server.loggers.app
        # The scheduler will run every "minutes" minutes. The minimum is 1.
        self.config = server.config.jobs
        self.minutes = self.config.minutes if self.config else None
        # Store, in attribute "last", the last minute corresponding to the last
        # scheduler execution. It allows to prevent several executions of the
        # same job(s) at the same minute.
        # ~~~
        self.last = time.localtime().tm_min
        # This attribute is initialised to the current minute. It means that, as
        # soon as the Appy server is started, the scheduler will not be able to
        # run until a minute elapses. This is very important in the context of
        # an automatic Appy server restart. Imagine a job that automatically
        # restarts the server. If the job takes less than one minute to execute,
        # the Appy server restarts, and, if "last" is empty, the job is executed
        # again... and again, as many times as it can within a minute.

    def runJobs(self):
        '''It's time to check if jobs must be run. Scan p_self.all and run
           jobs defined in it that must be run right now.'''
        jobs = self.config.all
        if not jobs: return
        now = time.localtime()
        for timeDef, job in jobs:
            if timeDef.mustRun(now):
                job.run(self)

    def scanJobs(self):
        '''This method is called by the server's infinite loop and checks
           whether jobs must be ran.'''
        # This method is called every "config.server.pollInterval" seconds. This
        # is too much: perform a real execution every "self.minutes" minutes.
        if not self.minutes: return
        current = time.localtime().tm_min
        if (current % self.minutes) or (current == self.last): return
        # Run jobs that must be run
        self.last = current
        self.runJobs()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import copy, types, re

from page import Page
from appy.px import Px
from appy import Object
from group import Group
from search import Search
from appy.gen import utils as gutils
from appy.shared import utils as sutils
from appy.gen.layout import Table, Layouts
from translations import FieldTranslations

# In this file, names "list" and "dict" refer to sub-modules. To use Python
# builtin types, use __builtins__['list'] and __builtins__['dict']

# ------------------------------------------------------------------------------
SEARCH_TRIGGER = '%s field %s:%s: executing search %s...'
SEARCH_DONE = 'fetched %d object(s).'

# ------------------------------------------------------------------------------
class Initiator:
    '''From a given field on a given object, it is possible to trigger the
       creation of another object. The most obvious example is the creation of a
       tied object via a Ref field with add=True. An "initiator" represents the
       (object, field) that initiates the object creation.

       In order to customize initiator behaviour, every Field sub-class can
       propose an Initiator sub-class, by overriding its static attribute
       "initiator".'''

    # Most initiators can be used as folders for the created objects
    asFolder = True

    def __init__(self, tool, req, info):
        self.tool = tool
        self.req = req
        # Extract the initiator object and field from p_info, parsed from a
        # "nav" key in the request, or simply already there as a tuple
        # (obj, field).
        if isinstance(info, basestring):
            self.info = info.split('.')
            self.obj = tool.getObject(self.info[0])
            self.field = self.extractField(req)
        else:
            self.obj, self.field = info
            self.info = None
        # After having created the object, if we are back from a popup, the
        # initiator may force to go back to some URL.
        self.backFromPopupUrl = None

    def extractField(self, req):
        '''Tries to get the initiator field from the request'''
        r = self.obj.getField(self.info[1])
        if r is None:
            # This can be an "objectless" initiator. In that case, p_self.obj is
            # the tool and the field's class is specified at another index in
            # p_self.info.
            r = self.obj.o.getAppyType(self.info[1], className=self.info[2])
        return r

    def getUrl(self):
        '''Returns the URL for going back to the initiator object, on the page
           showing self.field.'''
        return self.obj.o.getUrl(page=self.field.pageName, nav='no')

    def checkAllowed(self):
        '''Checks that adding an object via this initiator is allowed'''
        return True

    def updateParameters(self, params):
        '''Add the relevant parameters to the object edition page, related to
           this initiator.'''

    def goBack(self):
        '''Once the object has been created, where must we come back ? r_ can be
           - "view"        return to the "view" page of the created object;
           - "initiator"   return to the page, on self.obj, where self.field is.
        '''
        return 'view'

    def getNavInfo(self, new):
        '''If m_goBack is "view" and navigation-related information must be
           shown on the "view" page of the p_new object, this method returns
           it.'''

    def manage(self, new):
        '''Once the p_new object has been created, the initiator must surely
           perform a given action with it (ie, for a Ref field, the new object
           must be linked to the initiator object. This is the purpose of this
           method.'''

    def isComplete(self):
        '''We may have thought that an initiator was there, but once parsed, the
           "nav" key may reveal that there is no initiator at all. When it is
           the case, this method returns False and the initiator is considered
           to be inexistent.'''
        return True

# ------------------------------------------------------------------------------
class Field:
    '''Basic abstract class for defining any field'''
    Translations = FieldTranslations

    # Some global static variables
    nullValues = (None, '', [], {}, ())
    validatorTypes = (types.FunctionType, types.UnboundMethodType,
                      type(re.compile('')))
    labelTypes = ('label', 'descr', 'help')
    viewLayouts = ('view', 'cell')
    layoutTypes = ('view','edit','result','query','buttons','xml','sidebar')
    explicitLayoutTypes = ('buttons', 'sidebar')

    # Those attributes can be overridden by subclasses for defining,
    # respectively, names of CSS and Javascript files that are required by this
    # field, keyed by layoutType.
    cssFiles = {}
    jsFiles = {}

    # Some predefined layouts
    bLayouts = Table('lrv-f', width=None)
    dLayouts = 'lrv-d-f'
    # The *d*escription is visible, even on the *v*iew layout
    dvLayouts = {'edit': dLayouts, 'view': 'l-d-f'}
    hLayouts = 'lhrv-f'
    wLayouts = Table('lrv-f')
    # The following layouts are for fields within groups with style="grid"
    gLayouts = {'edit': 'frvl', 'search': 'l-f'}
    gdLayouts = {'edit': Table('frvl-d', width='99%'), 'search': 'l-f'}
    ghLayouts = {'edit': 'fhrvl', 'search': 'l-f'}
    gdhLayouts = {'edit': 'fhrvl-d', 'search': 'l-f'}
    # Unique, static cell layout
    cellLayout = Layouts.cell

    # For most fields, getting its value on some object is automatically done at
    # the framework level. Fields for which this must be disabled override the
    # following static attribute.
    customGetValue = False
    # Some fields are "outer": they are composed or inner fields
    outer = False
    # Parameters needed to re-trigger a search. Some fields applying on sets of
    # objects may need to re-trigger a search from these request parameters.
    searchParams = ('className', 'search', 'sortKey', 'sortOrder', 'filters')

    # The initiator class related to the Field
    initiator = Initiator

    # Render a field. Optional vars:
    # * fieldName   can be given as different as field.name for fields included
    #               in List fields: in this case, fieldName includes the row
    #               index.
    # * showChanges If True, a variant of the field showing successive changes
    #               made to it is shown.
    # * minimal     If True, the PX related to p_layoutType is directly called
    #               (pxView, pxCell, etc), instead of layout.pxRender. While
    #               this disables some functions like master/slave
    #               relationships, it returns a more lightweight code.
    pxRender = Px('''<x var="minimal=minimal|False;
      showChanges=showChanges|req.get('showChanges') == 'True';
      layoutType=layoutType|req.get('layoutType');
      isSearch=layoutType == 'search';
      hostLayout=req.get('hostLayout');
      name=fieldName or field.name|field.name;
      widgetName=isSearch and ('w_%s' % name) or name;
      rawValue=field.getValueIf(zobj, name, layoutType, disable=isSearch);
      value=not isSearch and \
            field.getFormattedValue(zobj,rawValue,layoutType,showChanges);
      requestValue=not isSearch and field.getStoredValue(zobj,name,True);
      inRequest=field.valueIsInRequest(zobj, req, name, layoutType);
      error=req.get('%s_error' % name) or req.validator.errors.get(name)|None;
      isMultiple=field.isMultiValued();
      masterCss=field.slaves and ('master_%s' % name) or '';
      tagCss=tagCss|None;
      tagCss=field.getTagCss(tagCss, layoutType);
      zobj=zobj or ztool;
      tagId='%s_%s' % (zobj.id, name);
      tagName=field.master and 'slave' or '';
      layoutTarget=field">:field.getPx(minimal, \
                                       hostLayout or layoutType, _ctx_)</x>''')

    def doRender(self, layoutType, obj, name=None, minimal=False, tagCss=None,
                 specific=None):
        '''Allows to call pxRender from code, to display the content of this
           field on p_obj in some specific context, for example in a Computed
           field.'''
        context = {'layoutType': layoutType, 'field': self, 'minimal': minimal,
                   'name': name or self.name, 'obj': obj, 'zobj': obj.o}
        if tagCss: context['tagCss'] = tagCss
        if specific: context.update(specific)
        # Get the current PX context: we will need to copy some entries from it.
        # If no such context exists, create one.
        req = obj.request
        if hasattr(req, 'pxContext'):
            ctx = req.pxContext
            Px.copyBaseVariables(context, ctx)
        else:
            ctx = Px.setBaseVariables(obj.tool, context, obj)
        # Call pxRender
        return self.pxRender(context).encode('utf-8')

    # Show the field content for some object on a list of referred objects
    pxRenderAsTied = Px('''
     <!-- The "title" field -->
     <x if="refField.name == 'title'">
      <x if="mayView">
       <x if="not field.menuUrlMethod or \
              (layoutType!='cell')">:field.pxObjectTitle</x>
       <x if="field.menuUrlMethod and (layoutType=='cell')"
          var2="info=field.getMenuUrl(zobj, \
           tied)">::tied.o.getListTitle(target=info[1], baseUrl=info[0])</x>
       <x if="not selector and tied.o.mayAct()">:field.pxObjectActions</x>
      </x>
      <div if="not mayView">
       <img src=":url('fake')" style="margin-right: 5px"/>
       <x>:_('unauthorized')</x></div>
     </x>
     <!-- Any other field -->
     <x if="(refField.name != 'title') and mayView">
      <x var="zobj=tied.o; obj=tied; layoutType='cell'; field=refField;
              fieldName=refField.name; minimal=not refField.inlineEdit"
         if="field.isShowable(zobj,'result')">:field.pxRender</x>
     </x>''')

    # Show the field content for some object on a list of results
    pxRenderAsResult = Px('''
     <!-- Title -->
     <x if="field.name=='title'"
        var2="navInfo=mode.getNavInfo(currentNumber); target=mode.target;
              backHook=backHook|None">
      <x if="mayView"
         var2="pageName=zobj.getDefaultViewPage();
               selectJs=inPopup and \
                        uiSearch.initiator.jsSelectOne(q, cbId) or ''">
       <x var="sup=zobj.getSupTitle(navInfo)" if="sup">::sup</x>
       <x>::zobj.getListTitle(mode=mode.titleMode, nav=navInfo, target=target, \
          page=pageName, inPopup=inPopup, selectJs=selectJs, highlight=True, \
          backHook=backHook)</x>
       <span style=":mode.showSubTitles and 'display:inline' or 'display:none'"
             name="subTitle" var="sub=zobj.getSubTitle()"
             class=":zobj.getCssFor('subTitle')"
             if="sub">::zobj.highlight(sub)</span>

       <!-- Actions -->
       <div if="not inPopup and uiSearch.showActions and zobj.mayAct()"
            class="objectActions" style=":'display:%s'% uiSearch.actionsDisplay"
            var2="layoutType='buttons';
                  editable=zobj.mayEdit();
                  locked=zobj.isLocked(user, 'main')">
        <!-- Edit -->
        <x if="editable">
         <a if="not locked"
            var2="linkInPopup=inPopup or (target.target != '_self')"
            target=":target.target"
            onclick=":target.getOnClick(backHook or zobj.id)"
            href=":zobj.getUrl(mode='edit', page=zobj.getDefaultEditPage(), \
                               nav=navInfo, inPopup=linkInPopup)">
          <img src=":url('edit')" title=":_('object_edit')"/>
         </a>
         <x if="locked" var2="lockStyle=''; page='main'">::obj.pxLock</x>
        </x>
        <!-- Delete -->
        <img if="not locked and zobj.mayDelete()" class="clickable"
             src=":url('delete')" title=":_('object_delete')"
             onClick=":'onDeleteObject(%s)' % q(zobj.id)"/>
        <!-- Fields (actions) defined with layout "buttons" -->
        <x if="not inPopup and (uiSearch.showActions == 'all')"
           var2="fields=zobj.getAppyTypes('buttons', 'main');
                 layoutType='cell'">
         <!-- Call pxCell and not pxRender to avoid having a table -->
         <x for="field in fields"
            var2="name=field.name; smallButtons=True">:field.pxCell</x>
        </x>
        <!-- Workflow transitions -->
        <x if="zobj.showTransitions('result')"
           var2="targetObj=zobj">:targetObj.appy().pxTransitions</x>
       </div>
      </x>
      <x if="not mayView">
       <img src=":url('fake')" style="margin-right: 5px"/>
       <x>:_('unauthorized')</x>
      </x>
     </x>
     <!-- Any other field -->
     <x if="(field.name != 'title') and mayView" var2="layoutType='cell'">
      <x if="field.isShowable(zobj, 'result')"
         var2="fieldName=field.name;
               minimal=not field.inlineEdit">:field.pxRender</x>
     </x>''')

    # Displays a field label
    pxLabel = Px('''<label if="field.hasLabel and field.renderLabel(layoutType)"
     lfor=":field.name">::_('label', field=field)</label>''')

    # Displays a field description
    pxDescription = Px('''<span if="field.hasDescr"
     class="discreet">::_('descr', field=field)</span>''')

    # Displays a field help
    pxHelp = Px('''<abbr title=":_('help', field=field)"><img
     src=":url('help')"/></abbr>''')

    # Displays validation-error-related info about a field
    pxValidation = Px('''<x><abbr if="error" title=":error"><img
     src=":url('warning')"/></abbr><img if="not error"
     src=":url('warning_no.gif')"/></x>''')

    # Displays the fact that a field is required
    pxRequired = Px('''<img src=":url('required.gif')"/>''')

    # Button for showing changes to the field
    pxChanges = Px('''
     <div if="zobj.hasHistory(name)" style="margin-bottom: 5px">
      <!-- Button for showing the field version containing changes -->
      <input if="not showChanges"
             var2="label=_('changes_show');
                   css=ztool.getButtonCss(label)" type="button" class=":css"
             value=":label" style=":url('changes', bg=True)"
             onclick=":'askField(%s,%s,%s,null,%s)' % \
                       (q(tagId), q(obj.url), q('view'), q('True'))"/>
      <!-- Button for showing the field version without changes -->
      <input if="showChanges"
             var2="label=_('changes_hide');
                  css=ztool.getButtonCss(label)" type="button" class=":css"
             value=":label" style=":url('changesNo', bg=True)"
             onclick=":'askField(%s,%s,%s,null,%s)' % \
                       (q(tagId), q(obj.url), q('view'), q('False'))"/>
     </div>''')

    # Widget for filtering object values on query results
    pxFilterText = Px('''
     <x var="name=field.name;
             filterId='%s_%s' % (mode.ajaxHookId, name);
             filterIdIcon='%s_icon' % filterId">
      <!-- Pressing the "enter" key in the field clicks the icon (onkeydown) -->
      <input type="text" size="7" id=":filterId"
             value=":mode.filters.get(name, '')"
             onkeydown=":'if (event.keyCode==13) document.getElementById ' \
                         '(%s).click()' % q(filterIdIcon)"/>
      <img id=":filterIdIcon" class="clickable" src=":url('funnel')"
           onclick=":'askBunchFiltered(%s,%s)' % (q(mode.ajaxHookId),q(name))"/>
     </x>''')

    def __init__(self, validator, multiplicity, default, defaultOnEdit, show,
      page, group, layouts, move, indexed, mustIndex, indexValue, searchable,
      specificReadPermission, specificWritePermission, width, height, maxChars,
      colspan, master, masterValue, focus, historized, mapping, generateLabel,
      label, sdefault, scolspan, swidth, sheight, persist, inlineEdit, view,
      cell, edit, xml, translations):
        # The validator restricts which values may be defined. It can be an
        # interval (1,None), a list of string values ['choice1', 'choice2'],
        # a regular expression, a custom function, a Selection instance, etc.
        self.validator = validator
        # Multiplicity is a 2-tuple indicating the minimum and maximum
        # occurrences of values.
        self.multiplicity = multiplicity
        # Is the field required or not ? (derived from multiplicity)
        self.required = self.multiplicity[0] > 0
        # Default value
        self.default = default
        # Default value on layout "edit". If None, self.default is used instead.
        self.defaultOnEdit = defaultOnEdit
        # Must the field be visible or not ?
        self.show = show
        # When displaying/editing the whole object, on what page and phase must
        # this field value appear?
        self.page = Page.get(page)
        self.pageName = self.page.name
        # Within self.page, in what group of fields must this one appear?
        self.group = Group.get(group)
        # The following attribute allows to move a field back to a previous
        # position (useful for moving fields above predefined ones).
        self.move = move
        # If indexed is True, a database index will be set on the field for
        # fast access.
        self.indexed = indexed
        # If "mustIndex", True by default, is specified, it must be a method
        # returning a boolean value. Indexation will only occur when this value
        # is True.
        self.mustIndex = mustIndex
        if not mustIndex and not callable(mustIndex):
            raise Exception('Value for param "mustIndex" must be a method.')
        # For an indexed field, the value stored in the index is deduced from
        # the field value and from the index type. If, for some reason, you want
        # to store another value, specify in parameter "indexValue" a method,
        # accepting the stored value as single arg, that can return an
        # alternative/transformed value to index.
        self.indexValue = indexValue
        # If specified "searchable", the field will be added to some global
        # index allowing to perform application-wide, keyword searches.
        self.searchable = searchable
        # Normally, permissions to read or write every attribute in a type are
        # granted if the user has the global permission to read or
        # edit instances of the whole type. If you want a given attribute
        # to be protected by specific permissions, set one or the 2 next boolean
        # values to "True". In this case, you will create a new "field-only"
        # read and/or write permission. If you need to protect several fields
        # with the same read/write permission, you can avoid defining one
        # specific permission for every field by specifying a "named"
        # permission (string) instead of assigning "True" to the following
        # arg(s). A named permission will be global to your application, so
        # take care to the naming convention. Typically, a named permission is
        # of the form: "<yourAppName>: Write|Read ---". If, for example, I want
        # to define, for my application "MedicalFolder" a specific permission
        # for a bunch of fields that can only be modified by a doctor, I can
        # define a permission "MedicalFolder: Write medical information" and
        # assign it to the "specificWritePermission" of every impacted field.
        self.specificReadPermission = specificReadPermission
        self.specificWritePermission = specificWritePermission
        # Widget width and height
        self.width = width
        self.height = height
        # While width and height refer to widget dimensions, maxChars hereafter
        # represents the maximum number of chars that a given input field may
        # accept (corresponds to HTML "maxlength" property). "None" means
        # "unlimited".
        self.maxChars = maxChars or ''
        # If the widget is in a group with multiple columns, the following
        # attribute specifies on how many columns to span the widget.
        self.colspan = colspan or 1
        # The list of slaves of this field, if it is a master
        self.slaves = []
        # The behaviour of this field may depend on another, "master" field
        self.setMaster(master, masterValue)
        # If a field must retain attention in a particular way, set focus=True.
        # It will be rendered in a special way.
        self.focus = focus
        # If we must keep track of changes performed on a field, "historized"
        # must be set to True.
        self.historized = historized
        # Mapping is a dict of contexts that, if specified, are given when
        # translating the label, descr or help related to this field.
        self.mapping = self.formatMapping(mapping)
        self.id = id(self)
        self.type = self.__class__.__name__
        self.pythonType = None # The True corresponding Python type
        # Get the layouts. Consult layout.py for more info about layouts.
        self.layouts = self.formatLayouts(layouts)
        # Can this field have values that can be edited and validated?
        self.validable = True
        # By default, if the base label for a field is in use in at least one of
        # its layouts (in p_self.layouts), it will be generated automatically in
        # your app's .pot and .po files. That being said, if the base label is
        # not used in any layout, but you still want it to be generated, for
        # using it in some other, non standard place like a pod or a custom
        # report, set the following attribute to True.
        self.generateLabel = generateLabel
        # If you want to avoid generating translation labels for this field, and
        # use instead translations already defined on another field, use the
        # following attribute "label". A label is made of 2 parts: the prefix,
        # based on class name, and the name, which is the field name by default.
        # ----------------------------------------------------------------------
        # If "label" is... | it will be understood as...
        # ----------------------------------------------------------------------
        # a string         | a new prefix = a new class name. Your field will
        #                  | use the label of an homonym field on another class.
        # ----------------------------------------------------------------------
        # a tuple          | a tuple (prefix, name), defining both a new prefix
        #                  | and a new field name. Your field will get the label
        #                  | of a field from another class, having another name.
        #                  | And if you want to reuse labels of another field
        #                  | defined on the same class, use a tuple of the form
        #                  |                 (None, name)
        # ----------------------------------------------------------------------
        self.label = label
        # When you specify a default value "for search" (= "sdefault"), on a
        # search screen, in the search field corresponding to this field, this
        # default value will be present.
        self.sdefault = sdefault
        # Colspan for rendering the search widget corresponding to this field.
        self.scolspan = scolspan or 1
        # Width and height for the search widget
        self.swidth = swidth or width
        self.sheight = sheight or height
        # "persist" indicates if field content must be stored in the database.
        # For some fields it is not wanted (ie, fields used only as masters to
        # update slave's selectable values). If you place a method in attribute
        # "persist", similarly, field value will not be stored, but your method
        # will receive it and you will do whatever you want with it.
        self.persist = persist
        # Can the field be inline-edited (on "view" or "cell" layouts)? A method
        # can be specified. If "inlineEdit" is or returns True, a click within
        # the field value as shown on the "view" or "cell" layout will switch
        # the widget to "edit" mode. If it returns 'icon', Switching to "edit"
        # will be done via an "edit" icon.
        self.inlineEdit = inlineEdit
        # If you want to use alternate PXs than Field.pxView, Field.pxCell and
        # Field.pxEdit, you can specify it in parameters "view", "cell" and
        # "edit". Instance attributes "pxView", "pxCell" and "pxEdit" below will
        # override their corresponding class attributes.
        if view is not None: self.pxView = view
        if cell is not None: self.pxCell = cell
        if edit is not None: self.pxEdit = edit
        # Standard marshallers are provided for converting values of this field
        # into XML. If you want to customize the marshalling process, you can
        # define a method in "xml" that will accept a field value and will
        # return a possibly different value. Be careful: do not return a chunk
        # of XML here! Simply return an alternate value, that will be
        # XML-marshalled.
        self.xml = xml
        # The PX for filtering field values. If None, it means that the field is
        # not filterable.
        self.filterPx = None
        # The standard system for getting field translations (label,
        # description, etc) is based on translations stored in persistent
        # Translation instances in tool.translations. But one may bypass this
        # system by placing an instance of class
        #            appy.fields.translations.FieldTranslations
        # in this attribute.
        self.translations = translations

    def init(self, name, klass, appName):
        '''When the application server starts, this secondary constructor is
           called for storing the name of the Appy field (p_name) and other
           attributes that are based on the name of the Appy p_klass, and the
           application name (p_appName).'''
        if hasattr(self, 'name'): return # Already initialized
        self.name = name
        # Determine prefix for this class
        if not klass: prefix = appName
        else:         prefix = gutils.getClassName(klass, appName)
        # Recompute the ID (and derived attributes) that may have changed if
        # we are in debug mode (because we recreate new Field instances).
        self.id = id(self)
        # Remember master name on every slave
        for slave in self.slaves: slave.masterName = name
        # Determine ids of i18n labels for this field, excepted if translations
        # are already present in self.tranlations.
        if not self.translations:
            labelName = name
            trPrefix = None
            if self.label:
                if isinstance(self.label, basestring): trPrefix = self.label
                else: # It is a tuple (trPrefix, name)
                    if self.label[1]: labelName = self.label[1]
                    if self.label[0]: trPrefix = self.label[0]
            if not trPrefix:
                trPrefix = prefix
            # Determine name to use for i18n
            self.labelId = '%s_%s' % (trPrefix, labelName)
            self.descrId = self.labelId + '_descr'
            self.helpId  = self.labelId + '_help'
        # Determine read and write permissions for this field
        rp = self.specificReadPermission
        if rp and not isinstance(rp, basestring):
            self.readPermission = '%s: Read %s %s' % (appName, prefix, name)
        elif rp and isinstance(rp, basestring):
            self.readPermission = rp
        else:
            self.readPermission = 'read'
        wp = self.specificWritePermission
        if wp and not isinstance(wp, basestring):
            self.writePermission = '%s: Write %s %s' % (appName, prefix, name)
        elif wp and isinstance(wp, basestring):
            self.writePermission = wp
        else:
            self.writePermission = 'write'
        if (self.type == 'Ref') and not self.isBack and self.back:
            # We must initialise the corresponding back reference
            self.back.klass = klass
            self.back.init(self.back.attribute, self.klass, appName)
        if self.type in ('List', 'Dict'):
            for subName, subField in self.fields:
                fullName = '%s_%s' % (name, subName)
                subField.init(fullName, klass, appName)
                subField.name = '%s*%s' % (name, subName)
        elif self.type == 'Switch':
            for case, fields in self.fields:
                for name, field in fields:
                    field.init(name, klass, appName)

    def __repr__(self):
        name = hasattr(self, 'name') and (':%s' % self.name) or ''
        return '<Field %s%s>' % (self.__class__.__name__, name)

    def reload(self, klass, obj):
        '''In debug mode, we want to reload layouts without restarting Zope.
           So this method will prepare a "new", reloaded version of p_self,
           that corresponds to p_self after a "reload" of its containing Python
           module has been performed.'''
        res = getattr(klass, self.name, None)
        if not res: return self
        if (self.type == 'Ref') and self.isBack: return self
        res.init(self.name, klass, obj.getProductConfig().PROJECTNAME)
        return res

    def isMultiValued(self):
        '''Does this type definition allow to define multiple values?'''
        maxOccurs = self.multiplicity[1]
        return (maxOccurs is None) or (maxOccurs > 1)

    def isSortable(self, usage):
        '''Can fields of this type be used for sorting purposes, when sorting
           search results (p_usage="search") or when sorting reference fields
           (p_usage="ref")?'''
        if usage == 'search':
            indexed = self.indexed
            return indexed and (indexed != 'TextIndex') and \
                   not self.isMultiValued()
        # For usage being "ref", this method will be overridden by fields for
        # which sort is allowed.

    def isShowable(self, obj, layoutType):
        '''When displaying p_obj on a given p_layoutType, must we show this
           field?'''
        # Check if the user has the permission to view or edit the field
        perm = (layoutType == 'edit') and self.writePermission or \
                                          self.readPermission
        if not obj.allows(perm): return
        # Evaluate self.show
        if callable(self.show):
            res = self.callMethod(obj, self.show)
        else:
            res = self.show
        # Take into account possible values 'view', 'edit', 'result'...
        if type(res) in sutils.sequenceTypes:
            for r in res:
                if r == layoutType: return True
            return
        elif res in self.layoutTypes:
            return res == layoutType
        # For showing a field on some layouts, they must be explicitly be
        # returned by the show method.
        if layoutType not in self.explicitLayoutTypes:
            return bool(res)

    def isRenderable(self, layoutType):
        '''In some contexts, computing m_isShowable can be a performance
           problem. For example, when showing fields of some object on layout
           "buttons", there are plenty of fields that simply can't be shown on
           this kind of layout: it is no worth computing m_isShowable for those
           fields. m_isRenderable is meant to define light conditions to
           determine, before calling m_isShowable, if some field has a chance to
           be shown or not.

           In other words, m_isRenderable defines a "structural" condition,
           independent of any object, while m_isShowable defines a contextual
           condition, depending on some object.'''
        # Most fields are not renderable on layout "buttons"
        if layoutType == 'buttons': return
        return True

    def isClientVisible(self, obj):
        '''This method returns True if this field is visible according to
           master/slave relationships.'''
        masterData = self.getMasterData()
        if not masterData: return True
        else:
            master, masterValue = masterData
            if masterValue and callable(masterValue): return True
            # Get the master value from the request
            rq = obj.REQUEST
            if not rq.has_key(master.name) and rq.has_key(self.name):
                # The master is not there: we cannot say if the slave must
                # be visible or not. But the slave is in the request. So we
                # should not prevent this value from being taken into account.
                return True
            reqValue = master.getRequestValue(obj)
            # reqValue can be a list or not
            if type(reqValue) not in sutils.sequenceTypes:
                return reqValue in masterValue
            else:
                for m in masterValue:
                    for r in reqValue:
                        if m == r: return True

    def inGrid(self, layout='edit'):
        '''Is this field in a group with style "grid" on p_layout?'''
        if not self.group: return
        if isinstance(self.group, __builtins__['dict']):
            for lay, group in self.group.iteritems():
                if lay != layout: continue
                return group and (group.style == 'grid')
            return
        return self.group.style == 'grid'

    def formatMapping(self, mapping):
        '''Creates a dict of mappings, one entry by label type (label, descr,
           help).'''
        if isinstance(mapping, __builtins__['dict']):
            # Is it a dict like {'label':..., 'descr':...}, or is it directly a
            # dict with a mapping?
            for k, v in mapping.iteritems():
                if (k not in self.labelTypes) or isinstance(v, basestring):
                    # It is already a mapping
                    return {'label':mapping, 'descr':mapping, 'help':mapping}
            # If we are here, we have {'label':..., 'descr':...}. Complete
            # it if necessary.
            for labelType in self.labelTypes:
                if labelType not in mapping:
                    mapping[labelType] = None # No mapping for this value
            return mapping
        else:
            # Mapping is a method that must be applied to any i18n message
            return {'label':mapping, 'descr':mapping, 'help':mapping}

    def formatLayouts(self, layouts):
        '''Standardizes the given p_layouts'''
        # First, get the layouts as a dictionary, if p_layouts is None or
        # expressed as a simple string.
        areDefault = False
        if not layouts: # Get the default ones as defined by the subclass
            layouts = self.computeDefaultLayouts()
            areDefault = True
        if isinstance(layouts, basestring):
            # This is a single layoutString, the "edit" one
            layouts = {'edit': layouts}
        elif isinstance(layouts, Table):
            # Idem, but with a Table instance
            layouts = {'edit': Table(other=layouts)}
        else:
            # Make a copy of the layouts, because every layout can be different,
            # even if the user decides to reuse one from one field to another.
            # This is because we modify every layout for adding
            # master/slave-related info, focus-related info, etc, which can be
            # different from one field to the other.
            layouts = copy.deepcopy(layouts)
            if 'edit' not in layouts:
                default = self.computeDefaultLayouts()
                if isinstance(default, __builtins__['dict']):
                    default = default['edit']
                layouts['edit'] = default
        # We have now a dict of layouts in p_layouts. Ensure now that a Table
        # instance is created for every layout (=value from the dict). Indeed,
        # a layout could have been expressed as a simple layout string.
        for layoutType in layouts.iterkeys():
            if isinstance(layouts[layoutType], basestring):
                layouts[layoutType] = Table(layouts[layoutType])
        # Derive "view" & "search" layouts from the "edit" layout when relevant
        if 'view' not in layouts:
            layouts['view'] = Table(other=layouts['edit'], derivedType='view')
        if 'search' not in layouts:
            layouts['search'] = Table(other=layouts['view'],
                                      derivedType='search')
        if self.focus:
            # We need to make it flashy
            layouts['view'].addCssClasses('focus')
            layouts['edit'].addCssClasses('focus')
        # If layouts are the default ones, set width=None instead of width=100%
        # for the field if it is not in a group (excepted for rich texts, refs
        # and calendars).
        if areDefault and not self.group and \
           not ((self.type == 'String') and (self.format == self.XHTML)) and \
           (self.type not in ('Ref', 'Calendar')):
            for layoutType in layouts.iterkeys():
                layouts[layoutType].width = ''
        # Remove letters "r" from the layouts if the field is not required
        if not self.required:
            for layoutType in layouts.iterkeys():
                layouts[layoutType].removeElement('r')
        # Derive some boolean values from the layouts
        self.hasLabel = self.hasLayoutElement('l', layouts)
        self.hasDescr = self.hasLayoutElement('d', layouts)
        self.hasHelp  = self.hasLayoutElement('h', layouts)
        return layouts

    def getPx(self, minimal, layoutType, context):
        '''Returns the PX corresponding to p_layoutType'''
        if minimal:
            # Call directly the layout-related PX on the field (bypass the
            # layout PX)
            r = getattr(self, 'px%s' % layoutType.capitalize())
        else:
            # Get the layout related to p_layoutType and render its PX
            if layoutType == 'cell':
                layout = self.cellLayout
            else:
                layout = self.layouts[layoutType]
            # Add the layout in the context, it is required by its PX
            context['layout'] = layout
            r = layout.pxRender
        return r

    def setMandatoriness(self, required):
        '''Updating mandatoriness for a field is not easy, this is why this
           method exists.'''
        # Update attributes "required" and "multiplicity"
        self.required = required
        if not required:
            minMult = 0
        else:
            minMult = 1
        self.multiplicity = (minMult, self.multiplicity[1])
        # Update the "edit" layout. We will not update editLayout.layoutString,
        # too tricky.
        editLayout = self.layouts['edit']
        if not required:
            # Remove pxRequired
            for row in editLayout.rows:
                for cell in row.cells:
                    for elem in cell.content:
                        if elem == 'pxRequired':
                            cell.content.remove(elem)
                            return
        else:
            # Add pxRequired
            content = editLayout.rows[0].cells[0].content
            content.insert(1, 'pxRequired')

    def setMaster(self, master, masterValue):
        '''Initialises the master and the master value if any'''
        self.master = master
        if master: self.master.slaves.append(self)
        # The semantics of attribute "masterValue" below is as follows:
        # - if "masterValue" is anything but a method, the field will be shown
        #   only when the master has this value, or one of it if multivalued;
        # - if "masterValue" is a method, the value(s) of the slave field will
        #   be returned by this method, depending on the master value(s) that
        #   are given to it, as its unique parameter.
        self.masterValue = gutils.initMasterValue(masterValue)

    def hasLayoutElement(self, element, layouts):
        '''This method returns True if the given layout p_element can be found
           at least once among the various p_layouts defined for this field.'''
        for layout in layouts.itervalues():
            if element in layout.layoutString: return True
        return False

    def getDefaultLayouts(self):
        '''Any subclass can define this for getting a specific set of default
           layouts. If None is returned, a global set of default layouts will be
           used.'''

    def getInputLayouts(self):
        '''Gets, as a string, the layouts as could have been specified as input
           value for the Field constructor.'''
        res = '{'
        for k, v in self.layouts.iteritems():
            res += '"%s":"%s",' % (k, v.layoutString)
        res += '}'
        return res

    def computeDefaultLayouts(self):
        '''This method gets the default layouts from a Field or the global
           default field layouts when they are not available.'''
        return self.getDefaultLayouts() or Layouts.defaults(self)

    def getCss(self, layoutType, res, config):
        '''This method completes the list p_res with the names of CSS files
           that are required for displaying widgets of self's type on a given
           p_layoutType. p_res is not a set because order of inclusion of CSS
           files may be important and may be loosed by using sets.'''
        if layoutType in self.cssFiles:
            for fileName in self.cssFiles[layoutType]:
                if fileName not in res:
                    res.append(fileName)

    def getJs(self, layoutType, res, config):
        '''This method completes the list p_res with the names of Javascript
           files that are required for displaying widgets of self's type on a
           given p_layoutType. p_res is not a set because order of inclusion of
           CSS files may be important and may be loosed by using sets.'''
        if layoutType in self.jsFiles:
            for fileName in self.jsFiles[layoutType]:
                if fileName not in res:
                    res.append(fileName)

    def isInner(self):
        '''Returns True if p_self is an inner field within a container field'''
        return '*' in self.name

    def getStoredValue(self, obj, name=None, fromRequest=False):
        '''Gets the value in its form as stored in the database, or in the
           request if p_fromRequest is True. It differs from calling
           m_getRequestValue because here, in the case of an inner field, the
           request value is searched within the outer value build and stored on
           the request.'''
        if self.isInner():
            # p_self is a sub-field into a List/Dict: p_name is of the form
            #            [outerName]*[name]*[rowId]
            outerName, name, rowId = name.split('*')
            if rowId == '-1': res = None # It is a template row
            else:
                # Get the outer value
                if fromRequest: res = obj.REQUEST.get(outerName, None)
                else:           res = getattr(obj.aq_base, outerName, None)
                # Access the inner value
                if res:
                    if rowId.isdigit():
                        rowId = int(rowId)
                        if rowId < len(res):
                            res = getattr(res[rowId], name, None)
                        else:
                            res = None
                    else:
                        res = res.get(rowId, None)
                        if res: res = res.get(name, None)
            # Return an empty string if fromRequest is True
            if fromRequest and (res is None): res = ''
        else:
            if fromRequest: res = self.getRequestValue(obj, self.name)
            else:           res = getattr(obj.aq_base, self.name, None)
        return res

    def getRequestValue(self, obj, requestName=None):
        '''Gets a value for this field as carried in the request object. In the
           simplest cases, the request value is a single value whose name in the
           request is the name of the field.

           Sometimes, several request values must be combined (ie: see the
           overriden method in the Date class).

           Sometimes (ie, a field within a List/Dict), the name of the request
           value(s) representing the field value does not correspond to the
           field name (ie: the request name includes information about
           the container field). In this case, p_requestName must be used for
           searching into the request, instead of the field name (self.name).'''
        name = requestName or self.name
        return obj.REQUEST.get(name, None)

    def setRequestValue(self, obj):
        '''Sets, in the request, field value on p_obj in its "request" form
           (=the way the value is carried in the request).'''
        # Get a copy of the field value on p_obj and put it in the request
        value = self.getCopyValue(obj)
        if value is not None:
            obj.REQUEST[self.name] = value

    def getRequestSuffix(self):
        '''In most cases, in the user interface, there is one homonym HTML
           element for representing this field. Sometimes, several HTML elements
           are used to represent one field (ie, for dates: one field for the
           year, one for the month, etc). In this case, every HTML element's
           name has the form <field name><suffix>. This method returns the
           suffix of the "main" HTML element.'''
        return ''

    def getValue(self, obj, name=None, layout=None):
        '''Gets, on_obj, the value for this field (p_self)'''
        # Possible values for parameters are described hereafter.
        # ----------------------------------------------------------------------
        # name    | p_name can be different from self.name if p_self is a
        #         | sub-field into a List: in this case, it includes the row
        #         | number.
        # ----------------------------------------------------------------------
        # layout  | In most cases, we don't care about the layout for getting
        #         | the value of p_self on p_obj. One exception is that a
        #         | default value can be defined specifically on layout "edit".
        #         | In that case, p_layout must be available in order to
        #         | determine the possible specific default value for this
        #         | layout.
        # ----------------------------------------------------------------------
        # Get the value from the database
        value = self.getStoredValue(obj, name)
        if self.isEmptyValue(obj, value):
            # If there is no value, get the default value if any. Determine
            # which one must be used: p_self.default or p_self.defaultOnEdit.
            if layout == 'edit':
                default = self.defaultOnEdit
                if default is None: default = self.default
            else:
                default = self.default
            # Get the default value, which can be a method
            if callable(default):
                try:
                    # Caching a default value can lead to problems. For example,
                    # the process of creating an object from another one, or
                    # from some data, sometimes consists in (a) creating an
                    # "empty" object, (b) initializing its values and
                    # (c) reindexing it. Default values are computed in (a),
                    # but it they depend on values set at (b), and are cached
                    # and indexed, (c) will get the wrong, cached value.
                    default = self.callMethod(obj, default, cache=False)
                except Exception:
                    # Already logged. Here I do not raise the exception,
                    # because it can be raised as the result of reindexing
                    # the object in situations that are not foreseen by
                    # method in self.default.
                    default = None
            return default
        return value

    def getValueIf(self, obj, name, layout, disable=False):
        '''Special method only called by the framework. For some fields (or
           according to some p_disable condition), value retrieval as performed
           here must be disabled.'''
        if disable or self.customGetValue: return
        return self.getValue(obj, name, layout)

    def getCopyValue(self, obj):
        '''Gets the value of this field on p_obj as with m_getValue above. But
           if this value is mutable, get a copy of it.'''
        return self.getValue(obj)

    def getComparableValue(self, obj):
        '''Get the value of this field on p_obj, as can be compared with its new
           version, coming from the validator.'''
        # For immutable values, it simply returns the currently stored value.
        # For mutable values, there are 2 cases:
        # (a) the value as produced by the validator completely overwrites the
        #     existing value. Thus, the previous and new versions can safely be
        #     compared, because being completely distinct. In this case, as for
        #     immutable values, the currently stored value can also be used
        #     as-is, as "comparable" value ;
        # (b) the value as produced by the validator updates the stored value.
        #     In this case, comparing the previous and new versions would lead
        #     to comparing the same value. Here, the "comparable" value must be
        #     a deep copy of the previous value and not the stored value itself.
        return self.getValue(obj)

    def getFormattedValue(self, obj, value, layoutType='view',
                          showChanges=False, language=None):
        '''p_value is a real p_obj(ect) value from a field from this type. This
           method returns a pretty, string-formatted version, for displaying
           purposes. Needs to be overridden by some child classes. If
           p_showChanges is True, the result must also include the changes that
           occurred on p_value across the ages. If the formatting implies
           translating some elements, p_language will be used if given, the
           user language else.'''
        if self.isEmptyValue(obj, value): return ''
        return value

    def getShownValue(self, obj, value, layoutType='view',
                      showChanges=False, language=None):
        '''Similar to m_getFormattedValue, but in some contexts, only a part of
           p_value must be shown. For example, sometimes we need to display only
           a language-specific part of a multilingual field (see overridden
           method in string.py).'''
        return self.getFormattedValue(obj,value,layoutType,showChanges,language)

    def getXmlValue(self, obj, value):
        '''This method allows a developer to customize the value that will be
           marshalled into XML. It makes use of attribute "xml".'''
        if not self.xml: return value
        return self.xml(obj, value)

    def getSearchValue(self, form, widgetName=None):
        '''Returns the search value (or interval of values) that has been
           encoded in a search p_form for matching values for this field.'''
        # The base search widget corresponding to p_self is prefixed with "w_".
        # p_widgetName can be specified to get the sub-value of another widget
        # that is part of the complete search widget for p_self.
        widgetName = widgetName or ('w_%s' % self.name)
        # p_value is the base (string) value encoded in the search form. But
        # the search value (or interval of values) can be made of several form
        # fields: overriden methods of this method in Field sub-classes may be
        # much more complex.
        r = form.get(widgetName)
        if isinstance(r, __builtins__['list']):
            # It is a list of values. Check if we have an operator for the
            # field, to see if we make an "and" or "or" for all those values.
            # "or" will be the default.
            operKey = 'o_%s' % self.name
            oper = ' %s ' % form.get(operKey, 'or').upper()
            r = oper.join(r)
        elif r:
            r = r.strip()
        return r or None

    def searchValueIsEmpty(self, form, widgetName=None):
        '''In some search p_form, can the search value or interval of values
           specified for this field (p_self) be considered as empty ?'''
        # If, for example, the search widget for this field is made of an
        # interval (from, to), p_widgetName can be given to correspond to the
        # widget part "to", while, by default, the specified part is "from". For
        # some Field sub-classes, the search widget may be even more complex and
        # made of more parts.
        widgetName = widgetName or ('w_%s' % self.name)
        # Conditions in overriden methods can be more complex. For example, if
        # an interval of values (min, max) is expected, specifying only "min" or
        # "max" will allow to perform a search. In this case, we will consider
        # the search value as empty if both "min" and "max" are not specified.
        value = form.get(widgetName)
        if value and isinstance(value, basestring): value = value.strip()
        return not value

    def getInlineEditableValue(self, obj, value, layoutType, name=None):
        '''Returns p_value as it must appear on a view layout, with code
           allowing to inline-edit it when relevant.'''
        # Disable inline-edition if not specified
        inlineEdit = self.getAttribute(obj, 'inlineEdit')
        if not inlineEdit: return value
        # Disable inline-edition if the user can't modify the value
        if not obj.allows(self.writePermission) \
           or obj.o.isLocked(obj.user, self.page.name):
            return value
        # Enable inline-edition, either via an additional icon or by clicking
        # in the value.
        hook = '%s_%s' % (obj.id, name or self.name)
        onClick = 'onclick="askField(\'%s\',\'%s\',\'edit:%s\')"' % \
                  (hook, obj.url, layoutType)
        if inlineEdit == 'icon':
            url = obj.tool.o.getIncludeUrl
            r = '<img src="%s" title="%s" class="inlineIcon" %s/>%s' % \
                (url('edit'), obj.translate('object_edit'), onClick, value)
        else:
            r = '<span class="editable" %s>%s</span>' % (onClick, value)
        return r

    def getFakeObject(self, value, request):
        '''Returns a fake object, to mimic an object storing this p_value for
           field p_self.'''
        r = Object()
        setattr(r, self.name, value)
        return Object(aq_base=r, REQUEST=request, id='0')

    def getIndexType(self):
        '''Returns the name of the technical, Zope-level index type for this
           field.'''
        # Normally, self.indexed contains a Boolean. If a string value is given,
        # we consider it to be an index type. It allows to bypass the standard
        # way to decide what index type must be used.
        if isinstance(self.indexed, str): return self.indexed
        if self.name == 'title': return 'TextIndex'
        return 'FieldIndex'

    def getIndexValue(self, obj, forSearch=False):
        '''This method returns a version for this field value on p_obj that is
           ready for indexing purposes. Needs to be overridden by some child
           classes.

           If p_forSearch is True, it will return a "string" version of the
           index value suitable for a global search.'''
        # Must we produce an index value?
        if not self.getAttribute(obj, 'mustIndex'): return
        # Start by getting the field value on p_obj
        res = self.getValue(obj)
        # Possibly transform the value
        if self.indexValue:
            res = self.indexValue(obj.appy(), res)
        # Zope catalog does not like unicode strings
        if isinstance(res, unicode): res = res.encode('utf-8')
        if forSearch and (res is not None):
            if type(res) in sutils.sequenceTypes:
                vals = []
                for v in res:
                    if isinstance(v, unicode): vals.append(v.encode('utf-8'))
                    else: vals.append(str(v))
                res = ' '.join(vals)
            else:
                res = str(res)
        return res

    def getIndexName(self, usage='search'):
        '''Gets the name of the Zope index that corresponds to this field.
           Indexes can be used for searching (p_usage="search"), filtering
           ("filter") or sorting ("sort"). The method returns None if the field
           named p_fieldName can't be used for p_usage.'''
        # Manage special cases
        if self.name == 'title':
            # For field 'title', Appy has a specific index 'SortableTitle',
            # because index 'Title' is a TextIndex (for searchability) and can't
            # be used for sorting.
            if usage == 'sort': return 'SortableTitle'
            elif usage == 'filter':
                return self.searchable and 'SearchableText' or 'Title'
            else: return 'Title'
        elif self.name == 'state': return 'State'
        elif self.name == 'SearchableText': return 'SearchableText'
        else:
            res = 'get%s%s'% (self.name[0].upper(), self.name[1:])
            if (usage == 'sort') and self.hasSortIndex(): res += '_sort'
        return res

    def hasSortIndex(self):
        '''Some fields have indexes that prevents sorting (ie, list indexes).
           Those fields may define a secondary index, specifically for sorting.
           This is the case of Ref fields for example.'''
        return

    def getSortValue(self, obj):
        '''Return the value of p_self on p_obj that must be used for sorting.
           While the raw p_value may be the value to use in most cases, it is
           not always true. For example, a string like "Gaëtan" could have
           "gaetan" as sort value.'''
        return self.getValue(obj)

    def getCatalogValue(self, obj, usage='search'):
        '''This method returns the index value that is currently stored in the
           catalog for this field on p_obj.'''
        if not self.indexed:
            raise Exception('Field %s: cannot retrieve catalog version of ' \
                            'unindexed field.' % self.name)
        return obj.getTool().getCatalogValue(obj, self.getIndexName(usage))

    def valueIsInRequest(self, obj, request, name=None, layoutType='view'):
        '''Is there a value corresponding to this field in the request? p_name
           can be different from self.name (ie, if it is a field within another
           (List) field). In most cases, checking that this p_name is in the
           request is sufficient. But in some cases it may be more complex, ie
           for string multilingual fields.'''
        return request.has_key(name or self.name)

    def getStorableValue(self, obj, value, complete=False):
        '''p_value is a valid value initially computed through calling
           m_getRequestValue. So, it is a valid representation of the field
           value coming from the request. This method computes the value
           (potentially converted or manipulated in some other way) as can be
           stored in the database.'''
        # More precisely, m_getStorableValue computes a value that can be used
        # as input for m_store. But this latter can further transform it. For
        # example, a value produced by p_getStorableValue for a Ref field can be
        # a list of object IDs; method Ref.store will then convert it to a list
        # of Appy objects and make cross-links between them and p_obj. If
        # p_complete is True, the value produced will be forced to be
        # as close as possible as the database value. In our Ref example, it
        # will be a list of Appy objects (but cross-links will not be
        # established).
        if self.isEmptyValue(obj, value): return
        return value

    def getInputValue(self, inRequest, requestValue, value):
        '''Gets the value that must be filled in the "input" widget
           corresponding to this field.'''
        if inRequest:
            return requestValue or ''
        else:
            return value or ''

    def isReadonly(self, obj):
        '''Returns True if, when this field is rendered on an edit layout as an
           input field, it must have attribute "readonly" set.'''
        return bool(self.getAttribute(obj, 'readonly'))

    def setSlave(self, slaveField, masterValue):
        '''Sets p_slaveField as slave of this field. Normally, master/slave
           relationships are defined when a slave field is defined. At this time
           you specify parameters "master" and "masterValue" for this field and
           that's all. This method is used to add a master/slave relationship
           that was not initially foreseen.'''
        slaveField.master = self
        slaveField.masterValue = gutils.initMasterValue(masterValue)
        if slaveField not in self.slaves:
            self.slaves.append(slaveField)
        # Master's init method may not have been called yet
        slaveField.masterName = getattr(self, 'name', None)

    def getMasterData(self):
        '''Gets the master of this field (and masterValue) or, recursively, of
           containing groups when relevant.'''
        if self.master: return self.master, self.masterValue
        group = self.getGroup('edit')
        if group: return group.getMasterData()

    def getMasterTag(self, layoutType):
        '''Generally, for a field, the name of the tag serving as master for
           driving slave fields is the name of the field itself. But sometimes
           it can be different (see Field sub-classes).'''
        return self.name

    def getTagCss(self, tagCss, layoutType):
        '''Gets the CSS class(es) that must apply to XHTML tag representing this
           field in the ui. p_tagCss may already give some base class(es).'''
        res = []
        isInner = self.isInner()
        if tagCss and not isInner:
            # For an inner tag, it has no sense to "inherit" from p_tagCss
            res.append(tagCss)
        # Add a special class when this field is the slave of another field
        if self.master:
            css = 'slave*%s*' % self.master.getMasterTag(layoutType)
            if not callable(self.masterValue):
                css += '*'.join(self.masterValue)
            else:
                css += '+'
            res.insert(0, css)
        # Define a base CSS class when the field is a sub-field in a List
        if isInner: res.append('no')
        return ' '.join(res)

    def getOnChange(self, zobj, layoutType, className=None):
        '''When this field is a master, this method computes the call to the
           Javascript function that will be called when its value changes (in
           order to update slaves).'''
        if not self.slaves: return ''
        q = zobj.getTool().quote
        # When the field is on a search screen, we need p_className
        cName = className and q(className) or 'null'
        return 'updateSlaves(this,null,%s,%s,%s,true)' % \
               (q(zobj.absolute_url()), q(layoutType), cName)

    def isEmptyValue(self, obj, value):
        '''Returns True if the p_value must be considered as an empty value.'''
        return value in self.nullValues

    def isCompleteValue(self, obj, value):
        '''Returns True if the p_value must be considered as "complete". While,
           in most cases, a "complete" value simply means a "non empty" value
           (see m_isEmptyValue above), in some special cases it is more subtle.
           For example, a multilingual string value is not empty as soon as a
           value is given for some language but will not be considered as
           complete while a value is missing for some language. Another example:
           a Date with the "hour" part required will not be considered as empty
           if the "day, month, year" part is present but will not be considered
           as complete without the "hour, minute" part.'''
        return not self.isEmptyValue(obj, value)

    def getCheckedInfo(self, req):
        '''Gets the status of checked elements from the p_req(uest)'''
        # It returns a tuple (ids, unchecked):
        # ----------------------------------------------------------------------
        #   "ids"     | is a list of object IDs;
        # "unchecked" | is True if these objects are unchecked, or False if
        #             | they are checked.
        # ----------------------------------------------------------------------
        ids = req['checkedUids'] and req['checkedUids'].split(',') or []
        unchecked = req['checkedSem'] == 'unchecked'
        return ids, unchecked

    def getSearchParams(self, req, layoutType):
        '''This method encodes in a string all the params in the request that
           are required for re-triggering a search.'''
        if not req.has_key('search') or (layoutType == 'cell'): return ''
        return ':'.join([req.get(key,'') for key in Field.searchParams])

    def getSearchResults(self, tool, searchParams):
        '''Execute the search whose parameters (p_searchParams) are returned by
           m_getSearchParams and return its results. Log the operation.'''
        # Get search parameters as local variables
        cmd = ', '.join(Field.searchParams)
        cmd += " = searchParams.split(':',%d)" % (len(Field.searchParams) - 1)
        exec cmd
        # Retrieve a Ref object when relevant
        refObject, refField = tool.getRefInfo()
        # Executing the query may take some time and potentially slow down
        # the system: log this action.
        tool.log(SEARCH_TRIGGER % (self.type, className, self.name, search))
        # (Re-)execute the query, but without any limit on the number of
        # results; return Appy objects.
        class_ = tool.getAppyClass(className)
        objs = tool.executeQuery(className, searchName=search, sortBy=sortKey,
                                 sortOrder=sortOrder, maxResults='NO_LIMIT',
                                 filters=tool.getFilters(class_, filters),
                                 refObject=refObject, refField=refField)
        r = [o.appy() for o in objs.objects]
        tool.log(SEARCH_DONE % len(r))
        return r

    def keepCheckedResults(self, req, objects):
        '''Among p_objects as retrieved via m_getSearchResults, keep only those
           being checked, according to m_getCheckedInfo.'''
        ids, unchecked = self.getCheckedInfo(req)
        i = len(objects) - 1
        # Remove, from search results, unchecked objects
        while i >= 0:
            if unchecked: remove = objects[i].id in ids
            else:         remove = objects[i].id not in ids
            if remove:
                del objects[i]
            i -= 1

    def validateValue(self, obj, value):
        '''This method may be overridden by child classes and will be called at
           the right moment by m_validate defined below for triggering
           type-specific validation. p_value is never empty.'''
        return

    def securityCheck(self, obj, value):
        '''This method performs some security checks on the p_value that
           represents user input.'''
        if not isinstance(value, basestring): return
        # Search Javascript code in the value (prevent XSS attacks)
        if '<script' in value:
            obj.log('Detected Javascript in user input.', type='error')
            raise Exception('Your behaviour is considered a security ' \
                            'attack. System administrator has been warned.')

    def validate(self, obj, value):
        '''This method checks that p_value, coming from the request (p_obj is
           being created or edited) and formatted through a call to
           m_getRequestValue defined above, is valid according to this type
           definition. If it is the case, None is returned. Else, a translated
           error message is returned.'''
        # If the value is required, check that a (complete) value is present
        if not self.isCompleteValue(obj, value):
            if self.required and self.isClientVisible(obj):
                # If the field is required, but not visible according to
                # master/slave relationships, we consider it not to be required.
                return obj.translate('field_required')
            else:
                return
        # Perform security checks on p_value
        self.securityCheck(obj, value)
        # Triggers the sub-class-specific validation for this value
        message = self.validateValue(obj, value)
        if message: return message
        # Evaluate the custom validator if one has been specified
        value = self.getStorableValue(obj, value)
        if self.validator and (type(self.validator) in self.validatorTypes):
            obj = obj.appy()
            if type(self.validator) != self.validatorTypes[-1]:
                # It is a custom function: execute it
                try:
                    validValue = self.validator(obj, value)
                    if isinstance(validValue, basestring) and validValue:
                        # Validation failed; and p_validValue contains an error
                        # message.
                        return validValue
                    else:
                        if not validValue:
                            return obj.translate('field_invalid')
                except Exception, e:
                    return str(e)
                except:
                    return obj.translate('field_invalid')
            else:
                # It is a regular expression
                if not self.validator.match(value):
                    return obj.translate('field_invalid')

    def store(self, obj, value):
        '''Stores, on p_obj, the p_value (produced by m_getStorableValue) that
           complies to p_self's type definition.'''
        persist = self.persist
        if not persist: return
        elif callable(persist):
            # Do not store the value but let this method do whatever she wants
            # with it.
            self.persist(obj.appy(), value)
        else:
            # Store the value on p_obj
            setattr(obj, self.name, value)

    def storeValueFromAjax(self, obj, value, previousData):
        '''Called by m_storeValueFromAjax to save p_value on p_obj'''
        self.store(obj, self.getStorableValue(obj, value))
        # Overridden methods may return some custom message part
        return ''

    def storeFromAjax(self, obj):
        '''Stores the new field value from an Ajax request, or do nothing if
           the action was canceled.'''
        rq = obj.REQUEST
        # Be the operation a "save" or "cancel", remove the lock on the page
        # (if is was set).
        obj.removeLock(field=self)
        # Do nothing more when canceling the ajax-save
        if rq.get('cancel') == 'True': return
        # Get the new value
        requestValue = rq['fieldContent']
        # Are we working on a normal or inner field ?
        if not self.isInner():
            # A normal field. Remember its previous value if it is historized.
            isHistorized = self.getAttribute(obj, 'historized')
            previousData = None
            if isHistorized: previousData = obj.rememberPreviousData(self)
            # Validate the value
            if self.validate(obj, requestValue):
                # Be minimalist: do nothing and return the previous value
                return
            # Store the new value on p_obj
            part = self.storeValueFromAjax(obj, requestValue, previousData)
            # Update the object history when relevant
            if isHistorized and previousData: obj.historizeData(previousData)
        else:
            # An inner field. No historization (yet) for it.
            part = ''
            # Retrieve the outer, inner fields and row number
            name, px = rq['px'].rsplit(':', 1)
            outer, inner, i = name.split('*')
            outer = obj.getAppyType(outer)
            # Update the value
            value = outer.getValue(obj)
            row = value[int(i)]
            setattr(row, inner, requestValue)
            # Store the complete value again on p_obj
            setattr(obj.aq_base, outer.name, value)
        # Update obj's last modification date
        from DateTime import DateTime
        obj.modified = DateTime()
        obj.reindex()
        obj.log('ajax-edited %s%s on %s.' % (self.name, part, obj.id))

    def callMethod(self, obj, method, cache=True):
        '''This method is used to call a p_method on p_obj. p_method is part of
           this type definition (ie a default method, the method of a Computed
           field, a method used for showing or not a field...). Normally, those
           methods are called without any arg. But one may need, within the
           method, to access the related field. This method tries to call
           p_method with no arg *or* with the field arg.'''
        obj = obj.appy()
        try:
            return gutils.callMethod(obj, method, cache=cache)
        except TypeError, te:
            # Try a version of the method that would accept self as an
            # additional parameter. In this case, we do not try to cache the
            # value (we do not call gutils.callMethod), because the value may
            # be different depending on the parameter.
            tb = sutils.Traceback.get()
            try:
                return method(obj, self)
            except Exception, e:
                obj.log('method %s:\n%s' % (method.func_name, tb), type='error')
                # Raise the initial error
                raise te
        except Exception, e:
            obj.log(sutils.Traceback.get(), type='error')
            raise e

    def getAttribute(self, obj, name, cache=True):
        '''Gets the value of attribute p_name on p_self, which can be a simple
           value or the result of a method call on p_obj.'''
        res = getattr(self, name)
        if not callable(res): return res
        return self.callMethod(obj, res, cache=cache)

    def getGroup(self, layoutType):
        '''Gets the group into wich this field is on p_layoutType'''
        if not self.group: return
        if isinstance(self.group, Group): return self.group
        # The group depends on p_layoutType
        if layoutType in self.group: return self.group[layoutType]

    def getGroups(self):
        '''Returns groups as a list'''
        res = []
        if not self.group: return res
        if isinstance(self.group, Group):
            res.append(self.group)
        else: # A dict
            for group in self.group.itervalues():
                if group and (group not in res):
                    res.append(group)
        return res

    def process(self, obj):
        '''This method is a general hook allowing a field to perform some
           processing after an URL on an object has been called, of the form
           <objUrl>/onProcess.'''
        return obj.goto(obj.absolute_url())

    def renderLabel(self, layoutType):
        '''Indicates if the existing label (if hasLabel is True) must be
           rendered by pxLabel. For example, if field is an action, the
           label will be rendered within the button, not by pxLabel.'''
        if not self.hasLabel: return
        # Label is always shown in search forms
        if layoutType == 'search': return True
        # If the field is within a "tabs" group, the label will already be
        # rendered in the corresponding tab. If the field is in a "grid" group,
        # the label is already rendered in a separate column.
        group = self.getGroup(layoutType)
        if group and (group.style in ('tabs', 'grid')): return
        return True

    def getSelectSize(self, isSearch, isMultiple):
        '''If this field must be rendered as a HTML "select" field, get the
           value of its "size" attribute. p_isSearch is True if the field is
           being rendered on a search screen, while p_isMultiple is True if
           several values must be selected.'''
        if not isMultiple: return 1
        prefix = isSearch and 's' or ''
        height = getattr(self, '%sheight' % prefix)
        if isinstance(height, int): return height
        # "height" can be defined as a string. In this case it is used to define
        # height via a attribute "style", not "size".
        return ''

    def getSelectStyle(self, isSearch, isMultiple):
        '''If this field must be rendered as a HTML "select" field, get the
           value of its "style" attribute. p_isSearch is True if the field is
           being rendered on a search screen, while p_isMultiple is True if
           several values must be selected.'''
        prefix = isSearch and 's' or ''
        # Compute CSS attributes
        res = []
        # Height
        height = getattr(self, '%sheight' % prefix)
        if isMultiple and isinstance(height, str):
            res.append('height: %s' % height)
            # If height is an integer value, it will be dumped in attribute
            # "size", not in CSS attribute "height".
        # Width
        width = getattr(self, '%swidth' % prefix)
        if isinstance(width, str):
            res.append('width: %s' % width)
            # If width is an integer value, it represents a number of chars
            # (usable for truncating the shown values), not a width for the CSS
            # attribute.
        return ';'.join(res)

    def getRadioStyle(self, isSearch=False):
        '''If this field must be rendered as radio buttons or checkboxes, and
           its height is specified as a string, restrict the visible part of the
           widget to this height, with a scrollbar when appropriate.'''
        prefix = isSearch and 's' or ''
        r = []
        height = getattr(self, '%sheight' % prefix)
        if isinstance(height, str):
            r.append('height:%s;overflow-y:auto' % height)
        width = getattr(self, '%swidth' % prefix)
        if isinstance(width, str):
            r.append('width:%s;overflow-x:auto' % width)
        return ';'.join(r)

    def getWidthInChars(self, isSearch):
        '''If attribute "width" contains an integer value, it contains the
           number of chars shown in this field (a kind of "logical" width). If
           it contains a string, we must convert the value expressed in px
           (or in another CSS-compliant unit), to a number of chars.'''
        prefix = isSearch and 's' or ''
        width = getattr(self, '%swidth' % prefix)
        if isinstance(width, int): return width
        if isinstance(width, str) and width.endswith('px'):
            return int(width[:-2]) / 5
        return 30 # Other units are currently not supported

    def getListHeader(self, ctx):
        '''When p_self is used as inner-field, within a table-like rendered
           container field, this method returns the content of the header row
           corresponding to this inner field.'''
        return ctx['_']('label', field=self)
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>Hello.</p>
<h2>Heading One</h2>
Blabla.<br />
<h3>SubHeading then.</h3>
Another blabla.<br /><br /><br /> '''
# I need a class.
class D:
    def getAt1(self):
        return xhtmlInput
dummy = D()



# ------------------------------------------------------------------------------
from appy.px import Px

# Input field for going to element number x. This PX is common to classes
# Siblings and Batch defined hereafter.
pxGotoNumber = Px('''
 <x var2="label=_('goto_number');
          gotoName='%s_%s_goto' % (obj.id, field.name);
          popup=inPopup and '1' or '0'">
  <span class="discreet" style="padding-left: 5px">:label</span> 
  <input type="text" size=":len(str(total)) or 1" onclick="this.select()"
         onkeydown=":'if (event.keyCode==13) document.getElementById' \
                     '(%s).click()' % q(gotoName)"/><img
         id=":gotoName" name=":gotoName"
         class="clickable" src=":url('gotoNumber')" title=":label"
         onclick=":'gotoTied(%s,%s,this.previousSibling,%s,%s)' % \
             (q(sourceUrl), q(field.name), total, q(popup))"/></x>''')

# ------------------------------------------------------------------------------
class Sibling:
    '''Represents a sibling element, accessible via navigation from
       another one.'''
    # Existing types of siblings
    types = ('previous', 'next', 'first', 'last')
    # Names of icons corresponding to sibling types
    icons = {'previous': 'arrowLeft',  'next': 'arrowRight',
             'first':    'arrowsLeft', 'last': 'arrowsRight'}

    def __init__(self, obj, type, nav, page, inPopup=False):
        # The sibling object
        self.obj = obj
        # The type of sibling
        self.type = type
        # The "nav" key
        self.nav = nav
        # The page that must be shown on the object
        self.page = page
        # Are we in a popup or not ?
        self.popup = inPopup and '1' or '0'

    def get(self, url, _):
        '''Get the HTML chunk allowing to navigate to this sibling'''
        js = "gotoSibling('%s/view','%s','%s','%s')" % \
             (self.obj.url, self.nav, self.page, self.popup)
        return '<img src="%s" title="%s" class="clickable" onclick="%s"/>' % \
               (url(Sibling.icons[self.type]), _('goto_%s' % self.type), js)

# ------------------------------------------------------------------------------
class Siblings:
    '''Abstract class containing information for navigating from one object to
       its siblings.'''

    # Icons for going to the current object's siblings
    pxGotoNumber = pxGotoNumber
    pxNavigate = Px('''
      <!-- Go to the source URL (search or referred object) -->
      <div if="not inPopup" class="goSource">::self.getGotoSource(url, _)</div>

      <!-- Show other navigation icons only when relevant -->
      <x if="self.total">
       <!-- Form used to navigate to any sibling -->
       <form name=":self.siblingsFormName" method="post" action=""
             style="display: inline">
        <input type="hidden" name="nav" value=""/>
        <input type="hidden" name="page" value=""/>
        <input type="hidden" name="popup" value=""/>
        <input type="hidden" name="criteria" value=""/>
       </form>

       <!-- Go to the first or previous page -->
       <x if="self.firstSibling">::self.firstSibling.get(url, _)</x>
       <x if="self.previousSibling">::self.previousSibling.get(url, _)</x>

       <!-- Explain which element is currently shown -->
       <span class="discreet"> 
        <x>:self.number</x> <b>//</b> <x>:self.total</x> </span>

       <!-- Go to the next or last page -->
       <x if="self.nextSibling">::self.nextSibling.get(url, _)</x>
       <x if="self.lastSibling">::self.lastSibling.get(url, _)</x>

       <!-- Go to the element number... -->
       <x if="self.showGotoNumber()"
          var2="field=self.field; sourceUrl=self.sourceObject.absolute_url();
                total=self.total"><br/><x>:self.pxGotoNumber</x></x></x>''',
     js='''
       function gotoCustomSearch() {
         // Post a form allowing to re-trigger a custom search
         var f = document.forms['gotoSource'];
         // Add search criteria from the browser's session storage
         f.criteria.value = sessionStorage.getItem(f.className.value);
         f.submit();
       }
       function gotoSibling(url, nav, page, popup) {
         var formName = (popup == '1') ? 'siblingsPopup' : 'siblings',
             f = document.forms[formName];
         // Navigate to the sibling by posting a form
         f.action = url;
         f.nav.value = nav;
         f.page.value = page;
         f.popup.value = popup;
         if (nav) {
           // For a custom search, get criteria from the session storage
           var elems = nav.split('.');
           if ((elems[0] == 'search') && (elems[2] == 'customSearch')) {
             f.criteria.value = sessionStorage.getItem(elems[1]);
           }
           // Avoid losing navigation
           f.action = f.action + '?nav=' + nav;
         }
         f.submit();
        }''')

    @staticmethod
    def get(nav, tool, inPopup):
        '''Analyse the navigation info p_nav and returns the corresponding
          concrete Siblings instance.'''
        elems = nav.split('.')
        Siblings = (elems[0] == 'ref') and RefSiblings or SearchSiblings
        return Siblings(tool, inPopup, *elems[1:])

    def computeStartNumber(self):
        '''Returns the start number of the batch where the current element
           lies.'''
        # First index starts at O, so we calibrate self.number
        number = self.number - 1
        batchSize = self.getBatchSize()
        res = 0
        while (res < self.total):
            if (number < res + batchSize): return res
            res += batchSize
        return res

    def __init__(self, tool, inPopup, number, total):
        self.tool = tool
        self.request = tool.REQUEST
        # Are we in a popup window or not ?
        self.inPopup = inPopup
        self.siblingsFormName = inPopup and 'siblingsPopup' or 'siblings'
        # The number of the current element
        self.number = int(number)
        # The total number of siblings
        self.total = int(total)
        # Do I need to navigate to first, previous, next and/or last sibling ?
        self.previousNeeded = False # Previous ?
        self.previousIndex = self.number - 2
        if (self.previousIndex > -1) and (self.total > self.previousIndex):
            self.previousNeeded = True
        self.nextNeeded = False     # Next ?
        self.nextIndex = self.number
        if self.nextIndex < self.total: self.nextNeeded = True
        self.firstNeeded = False    # First ?
        self.firstIndex = 0
        if self.previousIndex > 0: self.firstNeeded = True
        self.lastNeeded = False     # Last ?
        self.lastIndex = self.total - 1
        if (self.nextIndex < self.lastIndex): self.lastNeeded = True
        # Compute the IDs of the siblings of the current object
        self.siblings = self.getSiblings()
        # Compute the URL allowing to go back to the "source" = a given page of
        # query results or referred objects.
        self.sourceUrl = self.getSourceUrl()
        # Compute Sibling objects and store them in attributes named
        # "<siblingType>Sibling".
        nav = self.getNavKey()
        page = self.request.get('page', 'main')
        for siblingType in Sibling.types:
            exec 'needIt = self.%sNeeded' % siblingType
            name = '%sSibling' % siblingType
            setattr(self, name, None)
            if not needIt: continue
            exec 'index = self.%sIndex' % siblingType
            id = None
            try:
                # self.siblings can be a list (ref) or a dict (search)
                id = self.siblings[index]
            except KeyError: continue
            except IndexError: continue
            if not id: continue
            obj = self.tool.getObject(id, appy=True)
            if not obj: continue
            # Create the Sibling instance
            sibling = Sibling(obj, siblingType, nav % (index+1), page, inPopup)
            setattr(self, name, sibling)

    def getSuffixedBackText(self, _):
        '''Gets the p_backText, produced by m_getBackText (see sub-classes),
           suffixed with a standard text.'''
        text = self.getBackText()
        goto = _('goto_source')
        return text and ('%s - %s' % (text, goto)) or goto

    def getGotoSource(self, url, _):
        '''Get the link allowing to return to the source URL'''
        return '<a href="%s"><img src="%s"/>%s</a>' % \
               (self.sourceUrl, url('gotoSource'), self.getSuffixedBackText(_))

# ------------------------------------------------------------------------------
class RefSiblings(Siblings):
    '''Class containing information for navigating from one object to another
       within tied objects from a Ref field.'''
    prefix = 'ref'

    def __init__(self, tool, inPopup, sourceUid, fieldName, number, total):
        # The source object of the Ref field
        self.sourceObject = tool.getObject(sourceUid)
        # The Ref field in itself
        self.field = self.sourceObject.getAppyType(fieldName)
        # Call the base constructor
        Siblings.__init__(self, tool, inPopup, number, total)

    def getNavKey(self):
        '''Returns the general navigation key for navigating to another
           sibling.'''
        return self.field.getNavInfo(self.sourceObject, None, self.total)

    def getBackText(self):
        '''Computes the text to display when the user want to navigate back to
           the list of tied objects.'''
        _ = self.tool.translate
        return '%s - %s' % (self.sourceObject.Title(), _(self.field.labelId))

    def getBatchSize(self):
        '''Returns the maximum number of shown objects at a time for this
           ref.'''
        return self.field.maxPerPage

    def getSiblings(self):
        '''Returns the siblings of the current object'''
        return getattr(self.sourceObject, self.field.name, ())

    def getSourceUrl(self):
        '''Computes the URL allowing to go back to self.sourceObject's page
           where self.field lies and shows the list of tied objects, at the
           batch where the current object lies.'''
        # Allow to go back to the batch where the current object lies
        field = self.field
        startNumberKey = '%s_%s_objs_startNumber' % \
                         (self.sourceObject.id,field.name)
        startNumber = str(self.computeStartNumber())
        return self.sourceObject.getUrl(**{startNumberKey:startNumber,
                                           'page':field.pageName, 'nav':'no'})

    def showGotoNumber(self):
        '''Show "goto number" if the Ref field is numbered.'''
        return self.field.isNumbered(self.sourceObject)

# ------------------------------------------------------------------------------
class SearchSiblings(Siblings):
    '''Class containing information for navigating from one object to another
       within results of a search.'''
    prefix = 'search'

    def __init__(self, tool, inPopup, className, searchName, number, total):
        # The class determining the type of searched objects
        self.className = className
        # Get the search object
        self.searchName = searchName
        self.uiSearch = tool.getSearch(className, searchName, ui=True)
        self.search = self.uiSearch.search
        Siblings.__init__(self, tool, inPopup, number, total)

    def getNavKey(self):
        '''Returns the general navigation key for navigating to another
           sibling.'''
        return 'search.%s.%s.%%d.%d' % (self.className, self.searchName,
                                        self.total)

    def getBackText(self):
        '''Computes the text to display when the user want to navigate back to
           the list of searched objects.'''
        return self.uiSearch.translated

    def getBatchSize(self):
        '''Returns the maximum number of shown objects at a time for this
           search.'''
        return self.search.maxPerPage

    def getSiblings(self):
        '''Returns the siblings of the current object. For performance reasons,
           only a part of it is stored, in the session object.'''
        session = self.request.SESSION
        searchKey = self.search.getSessionKey(self.className)
        if session.has_key(searchKey): res = session[searchKey]
        else: res = {}
        if (self.previousNeeded and not res.has_key(self.previousIndex)) or \
           (self.nextNeeded and not res.has_key(self.nextIndex)):
            # The needed sibling UID is not in session. We will need to
            # retrigger the query by querying all objects surrounding this one.
            newStartNumber = (self.number-1) - (self.search.maxPerPage / 2)
            if newStartNumber < 0: newStartNumber = 0
            self.tool.executeQuery(self.className, search=self.search,
                                   startNumber=newStartNumber, remember=True)
            res = session[searchKey]
        # For the moment, for first and last, we get them only if we have them
        # in session.
        if not res.has_key(0): self.firstNeeded = False
        if not res.has_key(self.lastIndex): self.lastNeeded = False
        return res

    def getSourceUrl(self):
        '''Computes the (non-Ajax) URL allowing to go back to the search
           results, at the batch where the current object lies, or to the
           originating field if the search was triggered from a field.'''
        if ',' in self.searchName:
            # Go back to the originating field
            id, name, mode = self.searchName.split(',')
            obj = self.tool.getObject(id, appy=True)
            field = obj.getField(name)
            return '%s?page=%s' % (obj.url, field.page.name)
        else:
            url = '%s/query' % self.tool.absolute_url()
            # For a custom search, do not add URL params: we will build a form
            # and perform a POST request with search criteria.
            if self.searchName == 'customSearch': return url
            params = 'className=%s&search=%s&startNumber=%d' % \
                    (self.className, self.searchName, self.computeStartNumber())
            ref = self.request.get('ref', None)
            if ref: params += '&ref=%s' % ref
            return '%s?%s' % (url, params)

    def getGotoSource(self, url, _):
        '''Get the link or form allowing to return to the source URL'''
        if self.searchName != 'customSearch':
            return Siblings.getGotoSource(self, url, _)
        # For a custom search, post a form with the search criteria retrieved
        # from the browser's session.
        return '<form method="post" action="%s" name="gotoSource">' \
          '<input type="hidden" name="className" value="%s"/>' \
          '<input type="hidden" name="search" value="customSearch"/>' \
          '<input type="hidden" name="startNumber" value="%d"/>' \
          '<input type="hidden" name="ref" value="%s"/>' \
          '<input type="hidden" name="criteria" value=""/></form>' \
          '<a class="clickable" onclick="gotoCustomSearch()">' \
          '<img src="%s"/>%s</a>' % \
          (self.sourceUrl, self.className, self.computeStartNumber(),
           self.request.get('ref', ''), url('gotoSource'),
           self.getSuffixedBackText(_))

    def showGotoNumber(self): return

# ------------------------------------------------------------------------------
class Batch:
    '''Class for navigating between parts (=batches) within lists of objects'''
    def __init__(self, hook, total, length, size=30, start=0):
        # The ID of the DOM node containing the list of objects
        self.hook = hook
        # The total number of objects
        self.total = total
        # The effective number of objects in the current batch
        self.length = length
        # The maximum number of objects shown at once (in the batch). If p_size
        # is None, all objects are shown.
        self.size = size
        # The index of the first object in the current batch
        self.start = start

    def update(self, **kwargs):
        '''Update p_self's attributes with values from p_kwargs'''
        for name, value in kwargs.iteritems():
            setattr(self, name, value)

    def __repr__(self):
        return '<Batch hook=%s,start=%s,length=%s,size=%s,total=%s>' % \
               (self.hook, self.start, self.length, self.size, self.total) 

    # Icons for navigating among a list of objects: next, back, first, last...
    pxGotoNumber = pxGotoNumber
    pxNavigate = Px('''
     <div if="batch.total &gt; batch.size" align=":dright"
          var2="hook=q(batch.hook); size=q(batch.size)">

      <!-- Go to the first page -->
      <img if="(batch.start != 0) and (batch.start != batch.size)"
           class="clickable" src=":url('arrowsLeft')" title=":_('goto_first')"
           onclick=":'askBunch(%s,%s,%s)'% (hook, q(0), size)"/>

      <!-- Go to the previous page -->
      <img var="sNumber=batch.start - batch.size" if="batch.start != 0"
           class="clickable" src=":url('arrowLeft')" title=":_('goto_previous')"
           onclick=":'askBunch(%s,%s,%s)' % (hook, q(sNumber), size)"/>

      <!-- Explain which elements are currently shown -->
      <span class="discreet"> 
       <x>:batch.start + 1</x> <img src=":url('to')"/> 
       <x>:batch.start + batch.length</x> <b>//</b> <x>:batch.total</x>
      </span>

      <!-- Go to the next page -->
      <img var="sNumber=batch.start + batch.size" if="sNumber &lt; batch.total"
           class="clickable" src=":url('arrowRight')" title=":_('goto_next')"
           onclick=":'askBunch(%s,%s,%s)' % (hook, q(sNumber), size)"/>

      <!-- Go to the last page -->
      <img var="lastPageIsIncomplete=batch.total % batch.size;
                nbOfCompletePages=batch.total / batch.size;
                nbOfCountedPages=lastPageIsIncomplete and \
                                 nbOfCompletePages or nbOfCompletePages-1;
                sNumber= nbOfCountedPages * batch.size"
           if="(batch.start != sNumber) and \
               (batch.start != (sNumber-batch.size))" class="clickable"
           src=":url('arrowsRight')" title=":_('goto_last')"
           onclick=":'askBunch(%s,%s,%s)' % (hook, q(sNumber), size)"/>

      <!-- Go to the element number... -->
      <x var="gotoNumber=gotoNumber|False" if="gotoNumber"
         var2="sourceUrl=obj.url; total=batch.total">:batch.pxGotoNumber</x>
     </div>''')
# ------------------------------------------------------------------------------
